# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


__maintainer__ = 'Florian Boucault <florian@fluendo.com>'


from elisa.core.component import Component, ComponentError
from elisa.base_components.model import Model
from elisa.core.observers.observer import Observer
from elisa.core.observers.observable import Observable
from elisa.core import common
from elisa.core.input_event import *


class ModelNotSupported(Exception):
    pass


class Controller(Component, Observer, Observable):
    """
    A Controller is responsible for holding Models and extra information
    necessary for a View to render said Models. It also notify the
    Views about changes in the Models and in the extra information so
    that Views are always rendering the current content.

    A Controller is exposable on the network so that a remote View from
    another Elisa instance can render the data it holds.

    @ivar model:  Model associated to the Controller
    @type model: L{elisa.base_components.model.Model}
    """

    supported_models = ('base:model')

    default_associations = (
    ('base:model','base:controller'),
    )

    def __init__(self):
        Component.__init__(self)
        Observer.__init__(self)
        Observable.__init__(self)
        self._model = None
        self._backend = None
        self.parent = None
        self._focused = False
        self.loading = True


    def focus(self):
        """
        set the focus to the current controller
        """
        self.backend.focused_controller = self


    def focused__get(self):
        return self._focused


    def focused__set(self, value):
        self._focused = value
        self.focus_changed(value)


    def focus_changed(self, value):
        pass

    def backend__get(self):
        return self._backend


    def backend__set(self, new_backend):
        previous_backend, self._backend = self._backend, new_backend
        self.backend_changed(previous_backend, new_backend)


    def backend_changed(self, previous_backend, new_backend):
        pass


    def model__get(self):
        """
        @returns: the Model owned by the Controller
        @rtype:   Model
        """
        return self._model

    def attribute_set(self, key, value):
        # FIXME : find a better solution
        if key == 'loading':
            self.loading = value

    def model__set(self, model):
        """
        DOCME
        """
        if self._model == model:
            return

        if not model.path in self.supported_models:
            self.warning("controller %r does not support model %r" % (self.path, model.path))
            raise ModelNotSupported()

        self._model = model
        self.observe(model)
        self.model_changed()


    def model_changed(self):
        """
        This method is called by the model property when set.
        Controller can at this time securely initialize Model's
        properties.
        """
        pass

    def create_child_controller(self, model):
        """
        DOCME

        @raise TypeError: if controller is not a Controller object
        """

        if isinstance(model, Model):
            pr = common.application.plugin_registry
            controller_path_list = self.get_controller_paths(model.path)
            if len(controller_path_list) == 0:
                self.warning("can't create controller for model %r (with controller %r)" % (model.path, self.path))
                raise ComponentError(model.path)
            else:
                controller = pr.create_component(controller_path_list[0])
                controller.parent = self
                controller.backend = self.backend
                controller.model = model
                self.child_controller_created(controller)
                return controller
        else:
            self.warning("can't create controller for model %r (with controller %r)" % (model.path, self.path))
            raise TypeError


    def child_controller_created(self, controller):
        """
        called when the controller is created
        and the model is associated
        """
        pass


    def get_controller_paths(self, model_path):
        """
        DOCME
        """
        controller_paths = []
        for association in self.default_associations:
            if association[0] == model_path:
                controller_paths.append(association[1])
        return controller_paths


    def dispatch_input(self, input_event):
        """
        DOCME
        """
        if self.backend == None:
            # FIXME: how would that be possible to not have a backend ?
            self.handle_input(input_event)
        else:
            focused_child = self.backend.get_focused_child(self)
            if focused_child == None:
                return self.handle_input(input_event)
            else:
                notify_parent = focused_child.dispatch_input(input_event)
                if notify_parent:
                    return self.handle_input(input_event)
        return False

    def handle_input(self, input_event):
        """
        Translate an InputEvent into messages to the associated Controller.

        @param input_event:  the input event to translate
        @type input_event:   InputEvent
        @rtype : bool if True, input event are also posted to the parent
        """
        if input_event.action == EventAction.EXIT:
            common.application.stop()
            return False

        return True
