# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

"""
Module responsible for starting the Application
"""

__maintainer__ = 'Philippe Normand <philippe@fluendo.com>'

import pkg_resources
import sys, time
import os
import datetime

from elisa.core import common
common.boot()

from elisa.core import __version__, version_info, config, log
from elisa.core.utils import exception_hook
from elisa.core.bus import bus, bus_message

from elisa.core.utils import classinit
from elisa.core import thumbnailer
from elisa.core import media_uri
from elisa.core.utils.mime_getter import MimeGetter

from elisa.core import plugin_registry
from elisa.core import input_manager, media_manager, service_manager
from elisa.core import metadata_manager
from elisa.core import interface_controller, player_controller

from twisted.internet import reactor, defer, threads
from twisted.python import usage
from twisted.web import client

UPDATES_URL = "http://elisa.fluendo.com/updates/update.php?install_date=%(date)s&version=%(version)s"
CONFIG_DIR = os.path.join(os.path.expanduser('~'), ".elisa")
CONFIG_FILE = "elisa.conf"

DEFAULT_CONFIG = """\
[general]
version = '%(version)s'
install_date = '%(install_date)s'
media_providers = ['base:upnp_media', 'base:ipod_media', 'base:local_media', 'base:elisa_media', 'base:gnomevfs_media', 'base:audiocd']
metadata_providers = ['base:gst_metadata', 'base:taglib_metadata', 'base:cover_in_dir', 'base:cover_cache', 'base:amazon_covers']
service_providers = ['base:hal_service']
backends = ['backend1']
frontends = ['frontend1']

[media_scanner]
enabled = '1'
db_backend = 'sqlite'
database = 'elisa.db'
fivemin_location_updates = []
hourly_location_updates = []
daily_location_updates = []
weekly_location_updates = []
unmonitored_locations = []

[backend1]
activity = 'base:elisa_activity'
controller = 'poblenou:elisa_controller'
input_providers = ['base:lirc_input']

[frontend1]
backend = 'backend1'
view = 'poblenou:elisa_view'
theme = 'poblenou:tango_theme'
input_providers = ['pigment:pigment_input']

[base:main_menu_activity]
menu_activities = ['base:audio_activity', 'base:video_activity', 'base:image_activity', 'dvd:dvd_activity', 'service:service_activity']

[base:audio_activity]
# the audio media locations
locations = ['file://./sample_data/music/']

[base:image_activity]
# the picture media locations
locations = ['file://./sample_data/pictures/']

[base:video_activity]
# the video media locations
locations = ['file://./sample_data/movies/']

[service:service_activity]
# a list of activites, which should appear in the service_menu
service_activities = ['theme_switcher:theme_switcher_activity', 'service:about_activity']

[theme_switcher:theme_switcher_activity]
# a list of themes 'plugin:component' like 'classic:theme' 
themes = ['poblenou:tango_theme', 'poblenou:chris_theme', 'poblenou:poblenou_theme']

[poblenou:elisa_controller]
# wether or not the viewport should start on fullscreen state
start_fullscreen = '0'"""

class ConfigUpgrader(log.Loggable):

    def __init__(self, current_config, current_version, default_config):
        log.Loggable.__init__(self)
        self._current_config = current_config
        self._current_version = current_version
        self._default_config = config.Config('elisa_new.conf',
                                             default_config)
        
    def update_for(self, version):
        cfg = self._current_config
        if self._current_version is None and version[:2] == (0, 3):
            cfg = self._update_0_1_to_0_3()

        if cfg != self._current_config:
            full_path = os.path.join(cfg.get_config_dir(),
                                     self._current_config.get_filename())
            cfg.set_filename(full_path)
        else:
            self.info("No update needed")
        return cfg

    def _backup(self):
        cfg = self._current_config
        full_path = os.path.join(cfg.get_config_dir(), 'elisa.conf.bak')
        self.info("Doing a backup to the old config on %s", full_path)
        cfg.write(full_path)
    
    def _update_0_1_to_0_3(self):
        self.info('Upgrading from 0.1.x to 0.3.x')
        self._backup()
        cfg = self._default_config

        # media locations options have migrated to activities
        activity_map = {'plugins.music': 'base:audio_activity',
                        'plugins.movies': 'base:video_activity',
                        'plugins.pictures': 'base:image_activity'
                        }
        for old_section, new_section in activity_map.iteritems():
            locations = self._current_config.get_option('locations',
                                                        section=old_section,
                                                        default=[])
            new_locations = []
            location_labels = {}
            for location in locations:

                # meta:// URIs not supported anymore
                if location.startswith('meta'):
                    continue

                # FIXME: location/* not supported yet
                if location.endswith('*'):
                    location = location[:-1]

                location_uri = media_uri.MediaUri(location)

                # label parameter removed
                label = location_uri.get_param('label')
                if label:
                    location_uri.del_param('label')
                    location_labels[str(location_uri)] = label
                    
                new_locations.append(str(location_uri))
                
            cfg.set_option('locations', new_locations, section=new_section)
            activity_section = cfg.get_section(new_section)
            for location, label in location_labels.iteritems():
                activity_section[location] = {'label': str(label)}

        # start_fullscreen option moved to elisa_controller
        start_fullscreen = self._current_config.get_option('start_fullscreen',
                                                           default='0')
        controller_section = cfg.get_section('poblenou:elisa_controller')
        controller_section['start_fullscreen'] = str(start_fullscreen)

        # media_manager
        media_manager_section = self._current_config.get_section('media_manager')
        media_scanner_section = cfg.get_section('media_scanner')
        try:
            media_scanner_section['enabled'] = str(media_manager_section['enable_cache'])
        except KeyError:
            self.warning("enable_cache option not found in old config")

        try:
            media_scanner_section['database'] = media_manager_section['db_name']
        except KeyError:
            self.warning("db_name option not found in old config")
            
        for timebased in ('fivemin', 'hourly', 'daily', 'weekly'):
            opt_name = '%s_location_updates' % timebased
            try:
                media_scanner_section[opt_name] = media_manager_section[opt_name]
            except KeyError:
                self.warning("%s option not found in old config", opt_name)
                continue
        return cfg
                    
class Application(log.Loggable):
    """ Application is the entry point of Elisa. It groups all the necessary
    elements needed for Elisa to run. It is in charge of instanciating a
    Config and a PluginRegistry. Application also provides access to
    input events and data, and holds the user interfaces. It creates
    various managers (InputManager, MediaManager...),
    an InterfaceController and a DBBackend.

    @ivar plugin_registry: the PluginRegistry which loads and manages the plugins
    @type plugin_registry: L{elisa.core.plugin_registry.PluginRegistry}
    @ivar config:         Application's config file, storing options
    @type config:         L{elisa.core.config.Config}
    """

    # Allows property fget/fset/fdel/doc overriding
    __metaclass__ = classinit.ClassInitMeta
    __classinit__ = classinit.build_properties

    log_category = "application"

    def __init__(self, config_filename=None, show_tracebacks=False):
        """
        Application contructor. It:

         - loads global localization
         - loads the config file
         - loads the profiler

        @param config_filename: the config filename to use. Can be absolute
                                or relative path
        @type config_filename:  string or None to use default config file
        """
        log.Loggable.__init__(self)
        self.debug("Creating")

        self.show_tracebacks = show_tracebacks
        self.running = False
        self._plugin_registry = None

        self._config = None

        self._load_exception_hook()
        self._load_config(config_filename)
        self._load_profiler()
        
        common.set_application(self)

        self._plugin_registry = plugin_registry.PluginRegistry(self.config)
        self._plugin_registry.load_plugins()

        self._bus = bus.Bus()

        self._service_manager = service_manager.ServiceManager()
        self._metadata_manager = metadata_manager.MetadataManager()
        self._media_manager = media_manager.MediaManager(self.metadata_manager)
        self._input_manager = input_manager.InputManager()
        self._player_controller = player_controller.PlayerController()
        self._interface_controller = interface_controller.InterfaceController()
        self._thumbnailer = thumbnailer.Thumbnailer()
        self._mime_getter = MimeGetter()

    def _load_config(self, config_filename):
        if not config_filename:
            if not os.path.exists(CONFIG_DIR):
                os.makedirs(CONFIG_DIR)
            config_filename = os.path.join(CONFIG_DIR, CONFIG_FILE)

        self.info("Using config file: %r", config_filename)
        self._config_filename = config_filename
        today = datetime.date.today().isoformat()
        default_config = DEFAULT_CONFIG % {'version': __version__,
                                           'install_date': today}
        cfg = config.Config(config_filename, default_config=default_config)

        if not cfg.first_load:
            # ok we might have an old config format here
            version = cfg.get_option('version', section='general')
            upgrader = ConfigUpgrader(cfg, version, default_config)
            cfg = upgrader.update_for(version_info)
            
        self._install_date = cfg.get_option('install_date', section='general',
                                            default=today)
        self._config = cfg

    def _load_profiler(self):
        """
        This imports profiling modules (TODO)
        """
        enable = int(self.config.get_option('enable_profiling', default='0'))
        if enable:
            # TODO: import profiling modules and set things ready
            pass

    def _check_updates(self):
        url = UPDATES_URL % {'date': self._install_date, 'version': __version__}
        dfr = client.getPage(url)

        def got_result(result):
            # TODO: notify user about new version
            pass
        
        dfr.addCallback(got_result)
        
    def _load_exception_hook(self):
        """ Override the default system exception hook with our own
        """
        # FIXME: should this be in config?
        logdir = '/tmp'

        params = {'format': 'text', 'logdir': logdir,
                  'file': sys.stderr,
                  'display': self.show_tracebacks}
        self._except_hook = exception_hook.TracebackHook(**params)
        sys.excepthook = self._except_hook


    # FIXME: given the current use of this method, it should probably be called
    # something like log error to file
    def handle_traceback(self):
        """ Call this to force the exception hook to handle the exception on top
        of the stack. This can be handy to use from and except: block.
        """
        self._except_hook.handle()

    def plugin_registry__get(self):
        """
        PluginRegistry instance accessor.

        @returns: L{elisa.core.plugin_registry.PluginRegistry}
        """
        return self._plugin_registry

    def bus__get(self):
        """
        L{elisa.core.bus.bus.Bus}
        """
        return self._bus

    def metadata_manager__get(self):
        """ MetadataManager instance accessor

        @rtype: L{elisa.core.metadata_manager.MetadataManager}
        """
        return self._metadata_manager

    def mime_getter__get(self):
        """ MimeGetter instance accessor

        @rtype: l{elisa.core.utils.mime_getter.MimeGetter}
        """
        return self._mime_getter

    def service_manager__get(self):
        """ ServiceManager instance accessor

        @rtype: L{elisa.core.service_manager.ServiceManager}
        """
        return self._service_manager

    def player_controller__get(self):
        """ PlayerController instance accessor

        @rtype: L{elisa.core.player_controller.PlayerController}
        """
        return self._player_controller

    def interface_controller__get(self):
        """ InterfaceController instance accessor

        @rtype: L{elisa.core.interface_controller.InterfaceController}
        """
        return self._interface_controller

    def input_manager__get(self):
        """ InputManager instance accessor

        @rtype: L{elisa.core.input_manager.InputManager}
        """
        return self._input_manager

    def media_manager__get(self):
        """ MediaManager instance accessor

        @rtype: L{elisa.core.media_manager.MediaManager}
        """
        return self._media_manager

    def thumbnailer__get(self):
        """ Thumbnailer instance accessor

        @rtype: L{elisa.core.thumbnailer.Thumbnailer}
        """
        return self._thumbnailer

    def config__get(self):
        """ Config instance accessor

        @returns: L{elisa.core.config.Config}
        """
        return self._config

    def initialize(self):
        self._service_manager.initialize()
        self._metadata_manager.initialize()
        self._media_manager.initialize()
        self._input_manager.initialize()
        self._player_controller.initialize()
        self._interface_controller.initialize()

        self.bus.send_message(bus_message.ComponentsLoaded())

    def start(self):
        """ Execute the application. Start the Managers and the
        InterfaceController.
        """
        threads.deferToThread(self._check_updates)
        
        self.running = True
        self.info("Starting")
        self.bus.start()
        self.input_manager.start()
        self.media_manager.start(seconds=10,resume_scan=True)
        self.service_manager.start()
        self.interface_controller.start()


    def restart(self):
        return self.stop(restart=True)

    def stop(self, stop_reactor=True, restart=False):
        """ Close the application. Good idea to save the configuration
        here, since it's probable it has been updated.

        @param restart: tell the application to restart or not
        @type restart:  bool
        @rtype:         L{twisted.internet.defer.Deferred}
        """

        def stopped(r):
            self.player_controller.deinitialize()
            self.thumbnailer.stop()
            self.service_manager.stop()
            self.media_manager.stop()
            self.input_manager.stop()
            self.bus.stop()

            self.running = False

            if self.config:
                self.config.write()

            if restart:
                self.info("Restarting")
                self.running = True
                args = sys.argv
                args.insert(0, sys.executable)
                if reactor.running:
                    reactor.disconnectAll()
                    reactor.stop()
                os.spawnvpe(os.P_WAIT, sys.executable, args, os.environ)

            if stop_reactor and reactor.running:
                reactor.stop()

        if self.running:
            self.info("Stopping")
            dfr = self.interface_controller.stop()
            dfr.addCallback(stopped)
        else:
            dfr = defer.succeed(None)
            
        return dfr

class Options(usage.Options):
    """
    Application's command-line options definitions
    """

    optFlags = [['version', '', 'show elisa version'],
                ['twisted-version', '', 'show twisted version'],
                ['tracebacks', 't', 'display tracebacks'],
                ]

    tracebacks = False

    def parseArgs(self, config_file=None):
        self['config_file'] = config_file

    def opt_twisted_version(self):
        return usage.Options.opt_version(self)

    def opt_version(self):
        print 'Elisa version %s' % __version__
        sys.exit(0)


def main(args=None):
    """ Parse command-line options and start a new Application in the
    Twisted's reactor.
    """
    if not args:
        args = sys.argv

    options = Options()
    try:
        options.parseOptions(args[1:])
    except usage.UsageError, errortext:
        print '%s: %s' % (args[0], errortext)
        print '%s: Try --help for usage details.' % (args[0])
        sys.exit(1)

    try:
        app = Application(options['config_file'],
                          show_tracebacks=options['tracebacks'])
    except Exception, exc:
        raise
    else:
        app.initialize()
        app.start()
        reactor.run()
        app.stop()
        reactor.stop()
