# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


__maintainer__ = 'Philippe Normand <philippe@fluendo.com>'
__maintainer2__ = 'Florian Boucault <florian@fluendo.com>'

from elisa.core import common
from elisa.core import manager
from elisa.core import input_event
from elisa.core.utils import signal, classinit
from elisa.base_components.input_provider import PushInputProvider, PollInputProvider
from twisted.internet import reactor

class InputManager(manager.Manager):
    """
    InputManager provides a common place to retrieve input events
    coming from a GUI toolkit, additional input sources or even a
    network. InputEvents can be pushed by InputProviders or be polled
    by the InputManager, depending on the InputProvider implementation
    chosen (L{elisa.base_components.input_provider.PollInputProvider}
    or L{elisa.base_components.input_provider.PushInputProvider}.

    InterfaceController instances can connect to the manager's signals
    that are emitted when InputEvents are received.

    Work In Progress.

    @ivar key_pressed_signal:  signal triggered when a key has been pressed
    @type key_pressed_signal:  L{elisa.core.utils.signal.Signal}

    @ivar key_released_signal: signal triggered when a key has been released
    @type key_released_signal: L{elisa.core.utils.signal.Signal}

    """

##     key_pressed_signal = signal.Signal('key-pressed', input_event.InputEvent)
##     key_released_signal = signal.Signal('key-released', input_event.InputEvent)

##     button_pressed_signal = signal.Signal('button-pressed')
##     button_released_signal = signal.Signal('button-released')
##     button_scroll_signal = signal.Signal('button-scroll')

##     delete_signal = signal.Signal('delete')

    def __init__(self):
        """
        Create an InputManager instance
        """
        manager.Manager.__init__(self)

        # 25Hz refresh rate
        # FIXME: I would put at least 1/35 here due to the ability
        # of the human to detect sound delays above 30ms (average)
        # in the case you want to play a sound when a key is pressed.

        self._refresh_rate = 1 / 25.
        self._polling = False
        self._poll_call = None

    def start(self):
        """ Start a polling loop
        """
        self.info("Starting")
        self._polling = True
        # start a looping callLater
        self._poll_call = reactor.callLater(self._refresh_rate,
                                            self._poll_events)

    def stop(self):
        """ Clean the registered InputProviders and stop the manager's loop
        """
        self.info("Stopping")
        if self._poll_call:
            try:
                self._poll_call.cancel()
            except:
                pass
        self._polling = False
        manager.Manager.stop(self)

    def _poll_events(self):
        """ Poll each registered PollInputProvider for InputEvents to
        process.
        """
        input_providers = filter(lambda p: isinstance(p, PollInputProvider),
                                 self._components)

        for input_provider in input_providers:
            events = input_provider.get_input_events()
            for event in events:
                self.process_event(event, input_provider.path)

        if self._polling:
            self._poll_call = reactor.callLater(self._refresh_rate,
                                                self._poll_events)

    def process_event(self, event, provider_path):
        """ Fire the signal corresponding to the event

        Each event type is mapped to a signal instance to which other
        elisa components can connect (to e.g monitor user key press
        actions for instance).

        This method can be called by
        L{elisa.base_components.input_provider.PushInputProvider}
        components when they receive input data from the input device.

        @param event: the event to process
        @type event:  L{elisa.core.input_event.InputEvent}
        @param provider_path: the path of the InputProvider where the event is from
        @type provider_path:  str
        """
        self.debug("Event received: %r", event)

        if event == None:
            return

        self._emit_event(event, provider_path)

    def _emit_event(self, event, provider_path):
        """ Emit an event coming from a given InputProvider """
        sig_name = 'sig_%s' % provider_path
        try:
            sig = getattr(self, sig_name)
        except AttributeError:
            self.debug("Input provider signal %r not found", sig_name)
        else:
            sig.emit(event)

    def register_component(self, component):
        """ Register a new InputProvider in the Manager

        @param component: the InputProvider instance to register
        @type component:  L{elisa.base_components.input_provider.InputProvider}
        """
        manager.Manager.register_component(self, component)
        # register myself to push_input_providers
        if isinstance(component, PushInputProvider):
            component.input_manager = self
            component.bind()
        sig_name = 'sig_%s' % component.path
        setattr(self, sig_name, signal.Signal(sig_name,
                                              input_event.InputEvent))

    def unregister_component(self, component):
        """ Clean and InputProvider and unregister it from the Manager.

        @param component: the InputProvider instance to unregister
        @type component:  L{elisa.base_components.input_provider.InputProvider}
        """
        manager.Manager.unregister_component(self, component)
        component.clean()

    def subscribe(self, provider_path, callback):
        """
        DOCME
        """
        sig_name = 'sig_%s' % provider_path
        try:
            sig = getattr(self, sig_name)
        except AttributeError:
            msg = "Couldn't connect to input provider %r. Signal %r not found"
            self.debug(msg, provider_path, sig_name)
        else:
            sig.connect(callback)

    def unsubscribe(self, provider_path, callback):
        """
        DOCME
        """
        sig_name = 'sig_%s' % provider_path
        try:
            sig = getattr(self, sig_name)
        except AttributeError:
            self.debug("Couldn't disconnect from input provider %r. Signal %r not found", provider_path, sig_name)
        else:
            if sig:
                try:
                    sig.disconnect(callback)
                except Exception, exc:
                    self.debug(exc)
