# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

"""
URI parsing support
"""


__maintainer__ = 'Philippe Normand <philippe@fluendo.com>'
__maintainer2__ = 'Florian Boucault <florian@fluendo.com>'

import os, re, urllib
from elisa.core.utils import classinit

# Compiled regex used to parse URI.
URI_RE = None

class ParseException(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

def quote(data):
    """
    Make strings and dictionaries containing strings quoted using
    urllib.quote Return value has the same type as the parameter given
    to this function

    @param data: unquoted data which can contain unsafe characters like &=
    @type data: string or dict

    Attention: we are also quoting '/'! So don't use this EVER to a whole
    URL/URI
    """
    if isinstance(data, basestring):
        if isinstance(data, unicode):
            # urllib.quote doesn't support unicode objects very well
            data = data.encode('utf-8')
        quoted = urllib.quote(data,'')
    elif isinstance(data, dict):
        quoted = {}
        for k, v in data.iteritems():
            if isinstance(v, unicode):
                # urllib.quote doesn't support unicode objects very well
                v = v.encode('utf-8')
            quoted[k] = urllib.quote(v,'')
    elif isinstance(data, int):
        quoted = str(data)
    return quoted

def unquote(data):
    """
    This is just an proxy_method for urllib.unquote
    """
    ### BAHHH!!!!
    return unicode(urllib.unquote(data.encode('utf-8')).decode('utf-8'))


class MediaUri:
    """ Media URI management

    An URI is structured like this::

       scheme://[user:password@]host[:port]/path[/][?params][#fragment]

    This class is able to parse that and wrap access to all found
    attributes. When I'm parsing file:// URIs I'm allowed to replace
    paths like::

       ./path/to/foo

    With the following::

       "%s/path/to/foo" % os.curdir

    @ivar scheme:   the URI scheme
    @type scheme:   string
    @ivar user:     the optional username
    @type user:     string
    @ivar password: the optional password
    @type password: string
    @ivar host:     the URI hostname
    @type host:     string
    @ivar port:     URI optional port. Set to 0 if not found
    @type port:     int
    @ivar path:     the URI path, delimitted by slashes (/)
    @type path:     string
    @ivar fragment: optional URI fragment
    @type fragment: string
    @ivar params:   optional URI parameters
    @type params:   dict
    """

    # Allows property fget/fset/fdel/doc overriding
    __metaclass__ = classinit.ClassInitMeta
    __classinit__ = classinit.build_properties

    def __init__(self, data):
        """
        Create an Uri from various representations.

        Representation can be either a string containing the uri or
        the components of the uri stored in a dictionary.

        @param data:        An uri stored in a unicode string or described by
                            its components (scheme, user, password, host, port,
                            path, params, fragment), each of them being a string
        @type data:         string or dict
        @raises TypeError:  If none of location or parts has been provided
        @raises ParseException: If the location wasn't correctly parsed
        """
        location = None
        uri_parts = None

        self._label = u''

        if isinstance(data, MediaUri):
            # copycat, avoid another parse
            uri_parts = {'scheme': data.scheme, 'user': data.user,
                         'password': data.password, 'host': data.host,
                         'port': data.port, 'path': data.path,
                         'fragment': data.fragment, 'params': data._params}
            self._label = data._label
        elif isinstance(data, dict):
            # FIXME: scheme, host and path cannot be empty according to the docstring
            uri_parts = data
        elif isinstance(data, str):
            location = unicode(data)
        elif isinstance(data, unicode):
            location = data

        if uri_parts:
            # FIXME: scheme cannot be empty according to the docstring
            self.scheme = unicode(uri_parts['scheme'])
            self.user = unicode(uri_parts.get('user', ''))
            self.password = unicode(uri_parts.get('password', ''))
            # FIXME: host cannot be empty according to the docstring
            #self.host = unicode(uri_parts['host'])
            self.host = unicode(uri_parts.get('host', ''))
            # port is an integer, according to the docstring
            self.port = uri_parts.get('port', 0)

            self.fragment = unicode(uri_parts.get('fragment', ''))
            # FIXME: convert params to unicode
            self.params = uri_parts.get('params', {})

            path = uri_parts['path']

            # path should always start with / if it's not null
            # FIXME: why: Sorry, but this unusefull and not at is used in other cases. For e.g. the cdda-uri of gstreamer is: cdda://track-num. This is not working correctly...
            if path and not path.startswith('/'):
                path = "/%s" % path

            self.path = unicode(path)
        elif location:
            self._parse(location)
        else:
            raise TypeError("No location nor uri parts provided")

    def _uri_re(self):
        # compile the regex only once
        global URI_RE
        if not URI_RE:
            URI_RE = re.compile("^(([^:/?#]+):)?(//(([^:]+)\:([^@]*)@)?)?"
                                "([^?#/]*)([^?#]*)"
                                "(\?([^#]*))?(#(.*))?", re.UNICODE)
        return URI_RE

    def _parse(self, location):
        match = self._uri_re().search(location)
        if match:
            port = 0
            fscheme, scheme, auth, user_pass, user, passwd, host, path, fparams, params, ffragment, fragment = match.groups()

            if scheme:
                self.scheme = scheme
            else:
                self.scheme = u''

            if user:
                self.user = user
            else:
                self.user = u''

            if passwd:
                self.password = passwd
            else:
                self.password = u''

            host_port = host.split(':')
            if len(host_port) == 2:
                host, port = host_port
                port = int(port)

            if host:
                self.host = host
            else:
                self.host = u''

            self.port = port

            if fragment:
                self.fragment = fragment
            else:
                self.fragment = u''

            if scheme == 'file':
                # file scheme case: should disallow user/passwd/host
                if self.host == '.':
                    # file://./ special case
                    self.host = u''
                    if path.startswith('/'):
                        path = path[1:]
                    path = os.path.join(os.getcwd(), path)

                elif self.host:
                    # transfer host to beginning of path
                    path = "%s%s" % (self.host, path)
                    self.host = u''

                if not path.startswith('/'):
                    # path should always start with /
                    path = "/%s" % path

            self.path = unicode(path)

            if params:
                key_values = params.split('&')

                # ?foo special case
                if key_values[0].find('=') == -1:
                    key_values = [ "%s=" % key_values[0], ]
                elif key_values[0].rfind('=') != key_values[0].find('='):
                    # special case: params comma separated instead of
                    # &-separated
                    key_values = params.split(',')

                try:
                    self.params = dict([map(unicode, i.split('='))
                                        for i in key_values])
                except:
                    msg = "URI parameters were not quoted in %s" % location
                    raise ParseException(msg)
            else:
                self.params = {}

        else:
            raise ParseException("URI not parseable: %s" % location)

    def _parse_pyparsing(self, location):
        # FIXME: this method is deprecated
        """
        Parses an URI and fills the appropriates fields

        @param location:   An uri stored in a unicode string
        @type location:    string

        @raises ParseException: if the URI is not syntaxically correct
        """

        # Characters that cannot be used in a domain name
        domain_chars = NotAny('.') + CharsNotIn('+()\'\\[]|`_^={}#~%$*!:/;,?<>@')
        # Characters that cannot be used in a path
        allowed_chars = CharsNotIn('?#@:') #Word(alphanums + '_-%./~\\')
        # Characters that cannot be used in a parameter
        param_chars = CharsNotIn('?#,=')

        # Check for mandatory scheme
        scheme = Word(alphas).setResultsName('scheme') + \
                 Literal('://').suppress()

        # Check for optional user, :, password and @
        user_pass = Optional(allowed_chars + Literal(':').suppress() + \
                             allowed_chars + \
                             Literal('@').suppress()).setResultsName('user_pass')
        # Check for optional domain, and if found optional : and numeric port
        domain_port = Optional(domain_chars + \
                               Optional(Literal(':').suppress() + \
                                        Word(nums))).setResultsName('domain_port')
        # Retrieve optional whole path
        path = Optional(((Literal('./') ^ Literal('/')) + \
                         Optional(allowed_chars))).setResultsName('path')

        # Definition of a parameter delimiter
        params_delim = (Literal('&') ^ Literal(',')).suppress()

        # Optional dict of parameters
        params_dict = Dict(ZeroOrMore(Group(param_chars + \
                                            Optional('=').suppress() + \
                                            Optional(param_chars) + \
                                            Optional(params_delim))))

        # Check for optional parameter(s)
        params = Optional(Literal('?').suppress() + \
                          params_dict).setResultsName('params')

        # Retrive optional fragment
        fragment = Optional(Literal('#').suppress() + \
                            allowed_chars).setResultsName('fragment')

        # Grammer definition
        grammar = scheme + user_pass + domain_port + Optional(path + params) + \
                  fragment

        try:
            res = grammar.parseString(location)

            #TODO: debug to remove
            """
            print res
            print 'scheme:', res.scheme
            print 'user_pass:', res.user_pass
            print 'domain_port:', res.domain_port
            print 'path:', res.path
            print 'params:', res.params
            print 'fragment:', res.fragment
            """

            self.scheme = res.scheme
            if len(res.user_pass):
                self.user = res.user_pass[0]
                if len(res.user_pass) > 1:
                    self.password = res.user_pass[1]
            if len(res.domain_port):
                self.host = res.domain_port[0]
                if len(res.domain_port) > 1:
                    self.port = int(res.domain_port[1])
            if len(res.path):
                if (res.path[0] == './'):
                    res.path[0] = os.getcwd() + '/'
                self.path = ''.join(res.path)
            self.params = res.params

            # This is a bit weird.. the type of res.params
            # changes depending if some params are found or not
            if self.params == '':
                self.params = {}
            else:
                self.params = res.params.asDict()

            if len(res.fragment):
                self.fragment = res.fragment[0]

            # Check for file:// scheme validity
            if self.scheme == 'file':
                if len(self.path):
                    if not self.path.startswith('/') and \
                       not self.path.startswith('./'):
                        raise ParseException("invalid file:// scheme")
        except ParseException, pe:
            raise pe

    def params__set(self, data):
        self._params = data

    def params__get(self):
        params = dict([(k, urllib.unquote(v))
                       for (k,v) in self._params.iteritems()])
        return params

    def set_params(self, values):
        for key, value in values.iteritems():
            self.set_param(key, value)

    def set_param(self, name, value):
        quoted = quote(value)
        self._params[name] = quoted

    def get_param(self, name, default=u''):
        return urllib.unquote(self._params.get(name, default))

    def del_param(self, name):
        if name in self._params:
            del self._params[name]

    def extension__get(self):
        """ Retrieve file extension of the URI path

        If no extension is found, return empty string

        @rtype: unicode
        """
        # FIXME: is os.path.splitext doing any encoding conversion ?
        dummy, ext = os.path.splitext(self.path)
        if ext:
            # strip '.' prefix
            ext = ext[1:].lower()
        return unicode(ext)

    def filename__get(self):
        """ Return the filename of the Uri.

        Returns last path component label parameter like this::

          uri://path/to/foo

        then 'foo' is returned.

        @rtype: unicode
        """
        filename = ''

        if self.path and self.path != '/':
            idx = self.path.rfind('/')
            if idx != -1:
                filename = self.path[idx+1:]

        return unicode(filename)

    def label__set(self, value):
        if isinstance(value, str):
            self._label = value.decode('utf-8')#unicode(value)
        elif isinstance(value, unicode):
            self._label = value

    def label__get(self):
        """ Return a displayable string designing the Uri.

        Returns last path component if the URI has no predetermined
        label instance attribute set.

        @rtype: unicode
        """
        label = self._label
        if not label:
            path = self.path

            if path == '/':
                label = unicode(path)
            else:
                if self.path.endswith('/'):
                    path = self.path[:-1]

                idx = path.rfind('/')
                if idx != -1:
                    label = unicode(unquote(path[idx+1:]))

        return label


    def parent__get(self):
        """ Return my parent URI

        If my URI is like this::

          uri://path/to/foo

        I return:

          uri://path/to/

        @rtype: L{elisa.core.media_uri.MediaUri}
        """
        p = self.path
        if p.endswith('/'):
            p = p[:-1]
        s = p.split('/')[:-1]
        p = '/'.join(s) + '/'

        path = unicode(p)

        uri_parts = {'scheme': self.scheme,
                     'user': self.user,
                     'password': self.password, 'host': self.host,
                     'port': self.port, 'path': path}
        uri = MediaUri(uri_parts)
        return uri


    def join(self, path):
        """ Append the given path to my path attribute

        @param path: the path to append at the end of my path
        @type path:  string
        @rtype:      L{elisa.core.media_uri.MediaUri}
        """
        new_path = self.path
        if not new_path.endswith('/'):
            new_path += '/'
        new_path += path

        new_path = unicode(new_path)

        uri_parts = {'scheme': self.scheme,
                     'user': self.user,
                     'password': self.password, 'host': self.host,
                     'port': self.port, 'path': new_path,
                     'fragment': self.fragment, 'params': self.params}

        uri = MediaUri(uri_parts)
        return uri

    def _to_unicode(self):
        """ Textual representation of the URI

        @rtype: string
        """
        ret = u'%s://' % self.scheme
        if self.user:
            ret += u'%s:%s@' % (self.user, self.password)
        if self.host:
            ret += self.host
            if self.port:
                ret += u':%s' % self.port

        ret += self.path
        if len(self._params):
            ret += u'?'
            if type(self._params) != dict:
                d = self._params.asDict()
            else:
                d = self._params
            ret += u'&'.join('='.join((k,v)) for k,v in d.iteritems())

        if self.fragment:
            ret += u'#%s' % self.fragment

        assert(isinstance(ret, unicode))
        return ret

    __unicode__ = _to_unicode

    def __str__(self):
        """ Byte string representation of the URI

        @rtype: string
        """
        r = self._to_unicode()
        r = r.encode('utf-8')
        return r

    __repr__ = __str__

    def __cmp__(self, other_uri):
        """ Compare myself with another uri.

        @see help of builtin cmp()
        @param other_uri: The URI I'm comparing myself with
        @type other_uri:  L{elisa.core.media_uri.MediaUri}
        @rtype:           int
        @raise TypeError: When trying to compare with non-MediaUri object
        """
        if not isinstance(other_uri, MediaUri):
            raise TypeError("You can't compare MediaURI with other kind of objects")
        return cmp(self._to_unicode(), other_uri._to_unicode())

    def __eq__(self, other_uri):
        """
        @raise TypeError: When trying to compare with non-MediaUri object
        """

        if other_uri is None:
            return False
        if not isinstance(other_uri, MediaUri):
            raise TypeError("You can't compare MediaURI with other kind of objects")
        return self._to_unicode() == other_uri._to_unicode()

    def __ne__(self, other_uri):
        """
        @raise TypeError: When trying to compare with non-MediaUri object
        """

        if not isinstance(other_uri, MediaUri):
            return True
        return self._to_unicode() != other_uri._to_unicode()

    def __nonzero__(self):
        return len(self.scheme) != 0

    def __contains__(self, component):
        return component in self._to_unicode()

    def __add__(self, path):
        return self.join(path)

    def endswith(self, character):
        return self._to_unicode().endswith(character)

    def __getslice__(self, i, j):
        return self._to_unicode()[i:j]

if __name__ == '__main__':
    import sys
    for uri in sys.argv[1:]:
        u = MediaUri(uri)
        print u
        assert str(u) == uri, "%r != %r" % (u, uri)
