# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.



__maintainer__ = 'Florian Boucault <florian@fluendo.com>'


from observable import Observable
from observer import Observer
from elisa.core.utils.threadsafe_list import ThreadsafeList

class ListObservable(Observable, ThreadsafeList):

    # FIXME: this class should be made thread-safe

    def __init__(self, *args, **kw):
        ThreadsafeList.__init__(self, *args, **kw)
        Observable.__init__(self, *args, **kw)

    def __hash__(self):
        return Observable.__hash__(self)

    def __repr__(self):
        return "<ListObservable %r>" % ThreadsafeList.__repr__(self)

    def __eq__(self, other):
        return id(self) == id(other)

    def append(self, element):
        position = len(self)
        ThreadsafeList.append(self, element)
        for observer in self._observers:
            try:
                observer.inserted([element], position)
            except Exception, e:
                self.warning("Observer '%r' could not insert element '%r' at"
                             " position %s: %s" %
                                (observer, element, position, e))

    def extend(self, elements):
        position = len(self)
        ThreadsafeList.extend(self, elements)
        for observer in self._observers:
            try:
                observer.inserted(elements, position)
            except Exception, e:
                self.warning("Observer '%r' could not insert elements '%r' at"
                             " position %s: %s" %
                                (observer, elements, position, e))

    def insert(self, position, element):
        ThreadsafeList.insert(self, position, element)
        for observer in self._observers:
            try:
                observer.inserted([element], position)
            except Exception, e:
                self.warning("Observer '%r' could not insert element '%r' at"
                             " position %s: %s" %
                                (observer, element, position, e))


    def pop(self, position=-1):
        ret = ThreadsafeList.pop(self, position)
        for observer in self._observers:
            try:
                observer.removed([ret], position)
            except Exception, e:
                self.warning("Observer '%r' could not remove element '%r' at"
                             " position %s: %s" %
                                (observer, ret, position, e))
        return ret

    def remove(self, element):
        position = self.index(element)
        for observer in self._observers:
            try:
                observer.removed([element], position)
            except Exception, e:
                self.warning("Observer '%r' could not remove element '%r' at"
                             " position %s: %s" %
                                (observer, element, position, e))
        ThreadsafeList.remove(self, element)

    def reverse(self):
        ThreadsafeList.reverse(self)
        for observer in self._observers:
            try:
                observer.dirtied()
            except Exception, e:
                self.warning("Observer '%r' could not dirtied: %s" %
                                (observer, e))

    def sort(self, *args, **kwargs):
        ThreadsafeList.sort(self, *args, **kwargs)
        for observer in self._observers:
            try:
               observer.dirtied()
            except Exception, e:
                self.warning("Observer '%r' could not dirtied: %s" %
                                (observer, e))


    def __iadd__(self, elements):
        position = len(self)
        ret = ThreadsafeList.__iadd__(self, elements)
        for observer in self._observers:
            try:
                observer.inserted(elements, position)
            except Exception, e:
                self.warning("Observer '%r' could not insert elements '%r' at"
                             " position %s: %s" %
                                (observer, elements, position, e))
        return ret

    def __imul__(self, mult):
        position = len(self)
        elements = self[0:len(self)]
        elements *= mult - 1
        for observer in self._observers:
            try:
                observer.inserted(elements, position)
            except Exception, e:
                self.warning("Observer '%r' could not insert elements '%r' at"
                             " position %s: %s" %
                                (observer, elements, position, e))
        return ThreadsafeList.__imul__(self, mult)

    def __setitem__(self, key, value):
        ThreadsafeList.__setitem__(self, key, value)
        for observer in self._observers:
            try:
               observer.modified(key, value)
            except Exception, e:
                self.warning("Observer '%r' could not modify '%s' to"
                             " value '%s': %s" %
                                (observer, key, value, e)) 

    def __delitem__(self, key):
        element = self[key]
        for observer in self._observers:
            try:
                observer.removed([element], key)
            except Exception, e:
                self.warning("Observer '%r' could not remove element '%s' with"
                             " key '%s': %s" %
                                (observer, element, key, e))
        ThreadsafeList.__delitem__(self, key)

    def __setslice__(self, i, j, sequence):
        removed_elements = self[i:j]
        ThreadsafeList.__setslice__(self, i, j, sequence)
        for observer in self._observers:
            try:
                observer.removed(removed_elements, i)
                observer.inserted(sequence, i)
            except Exception, e:
                self.warning("Observer '%r' could not remove elements '%s'"
                             " beginning at %s and insert the sequence '%s'"
                             " again at '%s': %s" %
                                (observer, removed_elements, i, sequence, i, e))

    def __delslice__(self, i, j):
        removed_elements = self[i:j]
        for observer in self._observers:
            try:
                observer.removed(removed_elements, i)
            except Exception, e:
                self.warning("Observer '%r' could not remove elements '%s' at"
                             " '%s': '%s'" % (observer, removed_elements, i, e))        
        ThreadsafeList.__delslice__(self, i, j)

class ListObserver(Observer):

    def inserted(self, elements, position):
        print "%s inserted at %s" % (elements, position)
        raise NotImplementedError("method inserted not implemented by %r" % type(self))

    def removed(self, elements, position):
        print "%s removed from %s" % (elements, position)
        raise NotImplementedError("method removed not implemented by %r" % type(self))

    def modified(self, key, value):
        print "%s set to %s" % (key, value)
        raise NotImplementedError("method modified not implemented by %r" % type(self))

    def dirtied(self):
        print "list dirty"
        raise NotImplementedError("method dirtied not implemented by %r" % type(self))

if __name__ == "__main__":

    l = ListObservable()
    o = ListObserver()

    o.observe(l)

    print l
    l.append(3)
    print l
    l.extend([44, 4, 78, 18])
    print l
    l.insert(0, 15)
    print l
    l.pop(0)
    print l
    l.remove(3)
    print l
    l.reverse()
    print l
    l.sort()
    print l
    l += [12, 29]
    print l
    l *= 4
    print l
    l[2] = 555
    print l
    del l[2]
    print l
    l[0:3] = []
    print l
    del l[:]
    print l
