# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.



__maintainer__ = 'Philippe Normand <philippe@fluendo.com>'
__maintainer2__ = 'Florian Boucault <florian@fluendo.com>'


from elisa.core import log, common
from elisa.core.utils import classinit, signal
from elisa.core.media_uri import MediaUri
from elisa.extern import enum

from elisa.core.bus import bus, bus_message

import pygst
pygst.require('0.10')
import gst

from twisted.internet import reactor, defer, threads
import os
import copy
import urllib, time

class PlayingStarted(bus_message.Message):
    def __init__(self, data):
        bus_message.Message.__init__(self)
        self.data = data

class PlayingStopped(bus_message.Message):
    def __init__(self, data):
        bus_message.Message.__init__(self)
        self.data = data


STATES = enum.Enum("PLAYING", "PAUSED", "STOPPED")

SUBTITLE_FORMATS = ("asc", "txt", "srt", "smi", "ssa")


class Player(log.Loggable):
    """ This class provides a high level API for multimedia playback

    The Player is able to play media given their URIs. It
    provides features like:

    - playback control: play/pause/seek/speed
    - status reporting: volume/playing_time/..
    - playing queue management: next/previous
    - audio/video sinks management
    - video subtitles management
    - audio visualization support (libvisual, goom,...)

    In future API additions:

    - crossfading between medias
    - multiple audio/video/subtitles tracks selection in the same media
    - video hue/contrast/brightness set/get/save

    """

    # Allows property fget/fset/fdel/doc overriding
    __metaclass__ = classinit.ClassInitMeta
    __classinit__ = classinit.build_properties

    # REVIEW: maybe we should move these to player_controller and pass the Player
    #         instance at signal emission
    first_frame_decoded = signal.Signal('first_frame_decoded')
    error_signal = signal.Signal('error', str)
    end_of_stream = signal.Signal('end_of_stream', MediaUri)
    loading = signal.Signal('loading', bool)



    def __init__(self, player_id=0):
        """
        @param player_id:       ID of the player
        @type player_id:        int
        """
        self._id = player_id

        self.log_category = 'player_%d' % self._id
        log.Loggable.__init__(self)

        self._pipeline = None
        self._current_uri = None
        # FIXME: not sure if autosink is a wise decision: fakesink might be
        self._audiosink = gst.element_factory_make('autoaudiosink')
        self._videosink = gst.element_factory_make('autovideosink')
        self._playback_speed = 1.0
        self._unmuted_volume = 1.0
        self._subtitle_uri = None
        self.subtitle_font = 'Sans Bold 24'
        self.visualization_name = ''
        self.visualization_enabled = True
        #self._playlist = None
        self._create_pipeline()
        self._state = None
        self.seek_time = 150
        self._last_seek_time = 0
        self._loading = False
        self._play_in_progress = False
        self.file_type = ''
        
    def audio_sink__get(self):
        """ Audio sink access

        If we're using playbin, return the value of the 'audio-sink'
        property. Either find the element in the pipeline nammed
        'audio-sink'

        @rtype: L{gst.BaseSink}
        """

        return self._audiosink

    def _pipeline_set_state(self, state):
        # FIXME: Because of a bug in gnomevfs set_state might block.
        # This workaround should be removed as soon as possible.

        reactor.callInThread(self._pipeline.set_state, state)

    def audio_sink__set(self, sink):
        """ Audio sink setter

        @type sink: L{gst.BaseSink}
        """
        self._audiosink = sink
        if self._pipeline:
            self._pipeline.set_property('audio-sink', sink)

    def video_sink__get(self):
        """ Video sink access

        If we're using playbin, return the value of the 'video-sink'
        property. Either find the element in the pipeline named
        'video-sink'

        @rtype: L{gst.BaseSink}
        """

        return self._videosink

    def video_sink__set(self, sink):
        """ Video sink setter

        @type sink: L{gst.BaseSink}
        """
        self._videosink = sink
        if self._pipeline:
            self._pipeline.set_property('video-sink', sink)

    def current_uri__get(self):
        """ URI currently being played

        If the player has started some multimedia playback I return
        the URI it's currently playing. If nothing is playing I return None.

        @rtype: L{elisa.core.media_uri.MediaUri} or None
        """
        return self._current_uri

    def load_uri(self, uri):
        """ Load the media identified by an uri

        @param uri: media URI to play
        @type uri:  L{elisa.core.media_uri.MediaUri}
        """
        
        if uri.scheme == 'dvd':
            # FIXME : DVD without FLUENDO plugin is not yet supported
            self.warning("DVD playback without FLUENDO plugin is not yet supported")
            return
        
        if uri == self._current_uri:
            self.seek_to_location(0)
        else:
            self._play_in_progress = True
            self.loading.emit(True)
            self._loading = True
            #reactor.callInThread(self._load_uri, uri)
            threads.deferToThread(self._load_uri, uri)
        
    def _load_uri(self, uri):
        media_manager = common.application.media_manager
        
        def got_subs(subs):
            if subs:
                self.debug("Available subtitles: %s", subs)
                first = subs[0]
                self.subtitle_uri = first
                self.debug("Subtitles file %s loaded", self.subtitle_uri)
            else:
                self.debug("No subtitles to load")

        def got_media_type(media_type):
            self.subtitle_uri = ''
            self.file_type = media_type['file_type']
            
            real_uri = media_manager.get_real_uri(uri)
            self._current_uri = real_uri
            
            self._pipeline.set_state(gst.STATE_READY)
            self._play_in_progress = False
            states = self._pipeline.get_state(3000 * gst.MSECOND)
            if states[1] == gst.STATE_READY:

                # add a buffer probe to detect arrival of the first frame
                pad = self._videosink.get_pad('sink')
                self._probe_id = pad.add_buffer_probe(self._buffer_probe)
                
                audio_pad = self._audiosink.get_pad('sink')
                self._audio_probe_id = audio_pad.add_buffer_probe(self._audio_buffer_probe)
                # quote the uri before passing it to playbin
                real_uri.path = urllib.quote(real_uri.path.encode('utf8'))
                for param, value in real_uri.params.iteritems():
                    real_uri.set_param(urllib.quote(param),
                                       urllib.quote(value))

                self.debug("Loading %r", str(real_uri))

                self._pipeline.set_property('uri', str(real_uri))

                self.play()
                if self.file_type == 'video':
                    dfr = self.get_available_subtitles(uri)
                    dfr.addCallback(got_subs)
                
            else:
                self._loading = False
                self.loading.emit(False)

        # FIXME : bug hack : need a media provider for dvd:// ??
        if uri.scheme == 'dvd':
            got_media_type({'file_type' : 'video'})
        else:
            dfr = media_manager.get_media_type(uri)
            dfr.addCallback(got_media_type)

    def _create_pipeline(self):
        self._pipeline = gst.element_factory_make('playbin', 'playbin')
        bus = self._pipeline.get_bus()
        bus.connect("message", self._bus_message_cb)
        bus.add_signal_watch()

        self._pipeline.set_property('video-sink', self._videosink)
        self._pipeline.set_property('audio-sink', self._audiosink)

    def _bus_message_cb(self, bus, message):
        # FIXME: handle audio/video codec not found
        if message.type == gst.MESSAGE_EOS:
            self._on_eos()
        elif message.type == gst.MESSAGE_ERROR:
            err, msg = message.parse_error()
            if self._loading == True:
                self._loading = False
                self.loading.emit(False)
            self.warning(msg)
            self.error_signal.emit(msg)
            self._pipeline_set_state(gst.STATE_READY)
        elif message.type == gst.MESSAGE_STATE_CHANGED:
            new_state, old_state, pending = message.parse_state_changed()
            self._state = new_state

    def _buffer_probe(self, pad, buffer):
        pad.remove_buffer_probe(self._probe_id)
        self.first_frame_decoded.emit()
        if self._loading == True:
            self._loading = False
            self.loading.emit(False)
        self.info("First buffer decoded")
        return True

    def _audio_buffer_probe(self, pad, buffer):
        pad.remove_buffer_probe(self._audio_probe_id)
        if self._loading == True:
            self._loading = False
            self.loading.emit(False)
        self.info("First audio buffer decoded")
        return True
    
    def _on_eos(self):
        self.end_of_stream.emit(self._current_uri)
        self.info("End of stream reached")

    def play(self):
        """ Set the player state to playing mode
        """
        if not self._play_in_progress:
            if self._pipeline:
                self.info("Starting to play %r", str(self._current_uri))
                common.application.bus.send_message(PlayingStarted(self._current_uri),
                                                    sender = self)
    
                self._pipeline_set_state(gst.STATE_PLAYING)
                self.info("Playing %r", str(self._current_uri))

    def pause(self):
        """ Set the player state to paused mode
        """
        if self._pipeline:
            if self.state == STATES.PLAYING:
                self.info("Pausing Player")
                self._pipeline_set_state(gst.STATE_PAUSED)

    def toggle_play(self):
        """ Depending on current player's state perform a pause or
        un-pause action.
        """
        if self.state == STATES.PLAYING:
            self.pause()
        else:
            self.play()

    def stop(self):
        """ Set the player state to stopped mode
        """
        if self._pipeline:
            self.info("Stopping Player")
            common.application.bus.send_message(PlayingStopped(self._current_uri),
                                                    sender = self)
            self._pipeline_set_state(gst.STATE_READY)

    def gst_position__get(self):
        """
        Returns the position of the currently playing context

        @rtype: gst.FORMAT_TIME
        """

        if self._state not in (gst.STATE_PLAYING, gst.STATE_PAUSED):
            return gst.CLOCK_TIME_NONE

        try:
            position, format = self._pipeline.query_position(gst.FORMAT_TIME)
        except:
            #self.warning("gst_position retrieval failed")
            position = gst.CLOCK_TIME_NONE

        return position


    def position__get(self):
        """
        Returns the position of the currently playing context

        @rtype: long in nanoseconds,
        """
        position = self.gst_position
        if position == gst.CLOCK_TIME_NONE:
            position = 0

        return position


    def gst_duration__get(self):
        """
        Returns the duration of the currently playing context

        @rtype: gst.FORMAT_TIME
        """
        if self._state not in (gst.STATE_PLAYING, gst.STATE_PAUSED, None):
            return gst.CLOCK_TIME_NONE

        try:
            duration, format = self._pipeline.query_duration(gst.FORMAT_TIME)
        except:
            #self.warning("gst_duration retrieval failed")
            duration = gst.CLOCK_TIME_NONE

        return duration

    def duration__get(self):
        """
        Returns the duration of the currently playing context

        @rtype: long in nanoseconds,
        """
        duration = self.gst_duration
        if duration == gst.CLOCK_TIME_NONE:
            duration = 0

        return duration

    def status__get(self):
        """ Retrieve status information about the playing media

        Fetch current position reached by the player in the media and
        the media length.

        @rtype: tuple: (float, float)
        """
        return (self.position, self.duration)

    def state__get(self):
        """ Fetch the current state of the player, can be one of:

         - STATES.PLAYING
         - STATES.PAUSED
         - STATES.STOPPED
        """
        if self._pipeline.get_state()[1] == gst.STATE_PLAYING:
            return STATES.PLAYING
        elif self._pipeline.get_state()[1] == gst.STATE_PAUSED:
            return STATES.PAUSED
        
        return STATES.STOPPED
    
    def can_seek__get(self):
        # FIXME seek works only with file : with http we have
        # dead lock due to a GnomeVFS bug
        if self._current_uri and self._current_uri.scheme == 'file' \
            and self._state in (gst.STATE_PLAYING, gst.STATE_PAUSED):
            return True
        return False

    def seek_to_location(self, location, speed=None):
        """
        Seeks to given location

        @param location:        location to seek to
        @type location:         gst.FORMAT_TIME
        """
        
        if not self.can_seek:
            return
        
        
        now = time.time() * 1000
        
        if now < self._last_seek_time + self.seek_time:
            return
        
        self._last_seek_time = now
        
        # FIXME: could be transfered to position__set
        if speed == None:
            speed = self._playback_speed
        if location == gst.CLOCK_TIME_NONE or not self._pipeline:
            return
        human_loc = gst.TIME_ARGS(location)
        self.info("Seeking to location : %s at speed %s" % (human_loc, speed))
        if speed > 0.:
            start = location
            end_type = gst.SEEK_TYPE_NONE
            end = 0
        else:
            start = 0
            end = location
            end_type = gst.SEEK_TYPE_SET

        event = self._pipeline.seek(speed, gst.FORMAT_TIME,
                                    gst.SEEK_FLAG_FLUSH | gst.SEEK_FLAG_KEY_UNIT,
                                    gst.SEEK_TYPE_SET, start,
                                    end_type, end)

        self._pipeline.get_state(150 * gst.MSECOND)

    def seek_forward(self, seconds_or_fraction):
        """ Seek the current playing item some seconds
        forward. Seconds are supplied as integers. If the resulting
        location is valid within the playing item's length, the seek is
        performed. Otherwise, nothing happens.

        @param seconds_or_fraction: the time delta or absolute fraction of
                                    the media's length to seek at
        @type seconds_or_fraction:  int or float
        """
        # FIXME: not really needed with position__set

        if not self.can_seek:
            return
        
        seconds = None
        fraction = None
        if type(seconds_or_fraction) == int:
            seconds = seconds_or_fraction
        else:
            fraction = seconds_or_fraction
            if fraction >= 1.:
                fraction = 0.99

        if seconds:
            self.info("Seeking %s seconds forward" % seconds)
            new_location = self.position + (gst.SECOND * seconds)
        elif fraction:
            self.info("Seeking to %s location" % fraction)
            new_location = self.duration * fraction

        if new_location <= self.duration and new_location > 0:
            self.seek_to_location(new_location)

    def seek_backward(self, seconds_or_fraction):
        """ Seek the current playing item some seconds
        backward. Seconds are supplied as integers. If the resulting
        location is valid, the seek is performed. Otherwise, nothing
        happens.

        @param seconds_or_fraction: the time delta or absolute fraction of
                                    the media's length to seek at
        @type seconds_or_fraction:  int or float
        """
        # FIXME: not really needed with position__set

        if not self.can_seek:
            return
        
        seconds = None
        fraction = None
        if type(seconds_or_fraction) == int:
            seconds = seconds_or_fraction
        else:
            fraction = seconds_or_fraction
            fraction = self.old_fraction - fraction
            if fraction < 0.:
                fraction = 0.

        if seconds:
            self.info("Seeking %s seconds backward" % seconds)
            new_location = self.position - (gst.SECOND * seconds)
        elif fraction:
            self.info("Seeking to %s location" % fraction)
            new_location = self.duration * fraction
            if new_location >= 0:
                self.seek_to_location(new_location)

    def playback_speed__set(self, speed):
        """ Playback speed setter

        Playback speed is a float value between -10 and 10.
        All values outside this range will be brought back to it.

        @param speed: the new playback speed to set
        @type speed:  float
        """
        speed = float(speed)

        if speed > 10.0:
            speed = 10.0
        if speed < -10.0:
            speed = -10.0

        self.info("Setting playback speed to %s" % speed)

        location = self.position
        self._playback_speed = speed
        # FIXME: should avoid seeking for that
        self.seek_to_location(location, speed)

    def playback_speed__get(self):
        """ Playback speed getter

        Playback speed is a float value between -10 and 10.

        @rtype: float
        """
        return self._playback_speed

    def volume__set(self, volume):
        """ Set the volume level between 0 and 10

        @param volume: the new volume value
        @type volume:  float between 0 and 10
        """
        self.info("Setting volume to %r", volume)
        self._pipeline.set_property('volume', volume)

    def volume__get(self):
        """ Get the volume level

        @rtype: float between 0 and 10
        """
        volume = self._pipeline.get_property('volume')
        return volume

    def muted__get(self):
        """ return Mute state of the player

        @rtype: bool
        """

        return self.volume == 0

    def muted__set(self, muted):
        volume = self.volume

        if muted and volume != 0:
            self._unmuted_volume = volume
            self.volume = 0
        elif not muted and volume == 0:
            self.volume = self._unmuted_volume

    def mute(self):
        """ Mute the sound of the player
        """
        self.muted = True

    def unmute(self):
        """ Un-Mute the sound of the player
        """
        self.muted = False

    def toggle_mute(self):
        """ Toggle mute state of the player
        """
        self.muted = not self.muted


    def subtitle_uri__set(self, subtitle_uri):
        """ Load a subtitle file for given media_uri

        If we're using playbin, set the 'suburi' property of the
        pipeline to point to subtitle_uri value.

        If subtitle_uri is None the 'suburi' property is cleared.

        @param subtitle_uri: the subtitle uri
        @type subtitle_uri:  L{elisa.core.media_uri.MediaUri}
        """
        if self._pipeline:
            self._pipeline.set_property('suburi', subtitle_uri)
            self._pipeline.set_property('subtitle-font-desc',
                                        self.subtitle_font)
            self._subtitle_uri = subtitle_uri

    def subtitle_uri__get(self):
        """ Subtitle URI currently being used by the player

        @rtype: L{elisa.core.media_uri.MediaUri} or None
        """
        return self._subtitle_uri

    @defer.deferredGenerator
    def get_available_subtitles(self, uri):
        """ List available subtitles for the given media URI.

        This method can used the MediaProvider responsible to handle
        the given uri's scheme to scan the directory and find files
        starting like the media file but with different extensions
        (like: .sub, .srt, ...)

        @param uri: the media location to get subtitles for
        @type uri:  L{elisa.core.media_uri.MediaUri}
        @rtype:           L{twisted.internet.defer.Deferred}
        """
        subs = []
        if uri:
            self.debug("Searching subtitles for %s", uri)
            
            media_manager = common.application.media_manager
            filename = uri.filename
            ext_index = filename.index(uri.extension)
            basename = uri.filename[:ext_index-1]
            parent = uri.parent

            for ext in SUBTITLE_FORMATS:
                path = '%s.%s' % (basename, ext)
                sub_uri = parent.join(path)
                dfr = media_manager.is_directory(sub_uri)
                wfd = defer.waitForDeferred(dfr)
                yield wfd
                is_directory = wfd.getResult()
                if not is_directory:
                    handler = media_manager.blocking_open(sub_uri)
                    if handler:
                        handler.blocking_close()
                        subs.append(sub_uri)
                        
        yield subs

    def start_visualization(self):
        """ Show visualization animations for audio medias
        """
        if self._pipeline and self.visualization_enabled \
            and not self._pipeline.get_property('vis-plugin'):
            r = gst.registry_get_default()
            l = [x for x in r.get_feature_list(gst.ElementFactory) \
                 if (gst.ElementFactory.get_klass(x) == "Visualization")]
            e = [y for y in l if (y.get_name() == self.visualization_name)]
            if e:
                visplug = gst.element_factory_make(e.pop().get_name())
                self._pipeline.set_property('vis-plugin', visplug)


    def stop_visualization(self):
        """ Cancel visualization animations
        """
        if self._pipeline:
            self._pipeline.set_property('vis-plugin', None)


# PLAYLIST MANAGEMENT

#     def playlist__get(self):
#         """ Playlist access
#
#         @rtype: L{elisa.base_components.playlist.Playlist}
#         """
#         return self._playlist
#
#     def playlist__set(self, playlist):
#         """ Playlist setter
#
#         Update the playlist. If the player is already playing some
#         media it stops and start the new playlist from the beginning.
#
#         # Not sure it should stop. It could continue to play the current media and then use the new
#         playlist
#
#         @param playlist: the new playlist to set
#         @type playlist:  L{elisa.base_components.playlist.Playlist}
#         """
#         self._playlist = playlist
#         # FIXME: finish me
#
#     def add_uri(self, uri):
#         """ Add a new item at the end of the playing queue.
#
#         @param uri: the uri to add
#         @type uri: L{elisa.core.media_uri.MediaUri}
#         """
#         self.info("Adding new uri: %s" % uri)
#         if self._playlist:
#             self._playlist.add_uri(uri)
#         else:
#             # TODO: load_uri()
#             pass
#
#
#     def next(self):
#         """ Move forward to the next item in the playing queue. If the
#         end of playlist is reached, an IndexError is raised.
#
#         @rtype: bool
#         """
#         # NOTE: maybe IndexError should be raised by playlist.next() and
#         #       handled directly here
#         if self._playlist:
#             uri = self._playlist.next_uri()
#             if not uri:
#                 raise IndexError
#             else:
#                 return True
#         return False
#
#     def previous(self):
#         """ Move backward to the previous item in the playing
#         queue. If the beginning of the playlist is reached an IndexError
#         is raised.
#
#         @exception IndexError: when I'm trying to go before the first item of the list
#         @rtype:                bool
#         """
#         # NOTE: maybe IndexError should be raised by playlist.previous() and
#         #       handled directly here
#
#         if self._playlist:
#             uri = self._playlist.previous_uri()
#             if not uri:
#                 raise IndexError
#             else:
#                 return True
#         return False



if __name__ == '__main__':

    import sys
    import gobject
    import time

    if len(sys.argv) < 2:
        print "syntax: player.py URIS_OF_MEDIA"
        sys.exit(0)

    p = Player()

    def test_play():
        uris = [MediaUri(unicode(sys.argv[i])) for i in xrange(1,len(sys.argv))]

        p.load_uri(uris[0])
        p.play()
        while not p.state == STATES.PLAYING:
            time.sleep(1)
        p.seek_forward(100)
        p.playback_speed = 2
        time.sleep(2)
        p.muted = True
        time.sleep(1)



    loop = gobject.MainLoop()

    test_play()
    loop.run()
