# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

"""
Player instances access and global players control
"""

__author = 'Philippe Normand <philippe@fluendo.com>'
__maintainer__ = 'Philippe Normand <philippe@fluendo.com>'
__maintainer2__ = 'Florian Boucault <florian@fluendo.com>'


from elisa.core import log, player
from elisa.core.utils import classinit


class PlayerController(log.Loggable):
    """
    The PlayerController has responsability to create and hold
    L{elisa.core.player.Player} instances. Additionally it can perform
    global actions on all players at the same time.

    Player instances are identified by their name, which must be unique, this
    is the responsability of the developer. However there shouldn't be
    loads of Players, one "audio", one "video" and maybe one "ui-sounds"
    player.

    This is not a Manager because it doesn't act as a full blown proxy
    to players, when a component needs a Player, it requests it to the
    Controller which will create if it doesn't yet exist.

    @todo: API additions for new global actions

    @ivar _players: currently instantiated players
    @type _players: L{elisa.core.player.Player} list
    """

    # Allows property fget/fset/fdel/doc overriding
    __metaclass__ = classinit.ClassInitMeta
    __classinit__ = classinit.build_properties

    def __init__(self):
        log.Loggable.__init__(self)
        self._players = []
        # FIXME: this is _not_ cross-platform
        #self._mixer = gst.element_factory_make('alsamixer')
        self._id_counter = 0

    def initialize(self):
        """
        """
        self.info("Initializing")

    def deinitialize(self):
        """
        """

#   def volume__set(self, volume):
#       """ Set the volume level of the alsamixer element, between 0 and 100
#
#       @param volume: the new volume value
#       @type volume:  integer between 0 and 100
#       """
#
#       # FIXME: this is _not_ cross-platform
#       self._mixer.set_property('volume', volume)
#
#   def volume__get(self):
#       """ Get the volume level of the alsamixer element
#
#       @rtype: integer between 0 and 100
#       """
#
#       # FIXME: this is _not_ cross-platform
#       volume = self._mixer.get_property('volume')
#       return volume

    def create_player(self):
        """
        Returns a new player instance

        @rtype:      L{elisa.core.player.Player}
        """

        self.debug("Creating new player")
        self._id_counter += 1
        p = player.Player(self._id_counter)
        self._players.append(p)
        return p

    def remove_player(self, player_instance):
        """
        Remove player_instance Player

        @param player_instance: Player instance to remove
        @type player_instance: L{elisa.core.player.Player}
        """
        for i in self._players:
            if i == player_instance:
                self._players.remove(i)
                break


    def _global_action(self, method_name, *params):
        for player_instance in self._players:
            method = getattr(player_instance, method_name)
            method(*params)

    def play_all(self):
        """
        DOCME
        """
        self._global_action("play")

    def pause_all(self):
        """
        DOCME
        """
        self._global_action("pause")

    def stop_all(self):
        """ Stop all currently instantiated players
        """
        self._global_action("stop")

    def mute_all(self):
        """ Mute all players
        """
        self._global_action("mute")

    def unmute_all(self):
        """ Unmute all players
        """
        self._global_action("unmute")

    def toggle_mute_all(self):
        """ Toggle mute value of all players
        """
        self._global_action("toggle_mute")
