# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

"""
Plugin base class, the mother of all Elisa plugins
"""


__maintainer__ = 'Philippe Normand <philippe@fluendo.com>'

from elisa.core.utils import classinit, misc
from elisa.core import log
from elisa.core.config import Config
from elisa.extern import msgfmt
from elisa.core import component

import os
import gettext
import locale
import inspect

class Plugin(log.Loggable):
    """
    Plugin Base Class

    This class acts as a Component factory.

    @cvar components:       Component implementations provided by the Plugin
    @type components:       L{elisa.core.component.Component} list
    @cvar translation_path: plugin's .po directory
    @type translation_path: string
    @ivar name:             plugin's name. Should be unique
    @type name:             string

    @todo: versioning

    @note: if translation_path is not empty, .po files will be
    compiled and localization support will be installed. Plugins need
    to override local _() function in their sources to use
    localization. Please see docs/plugins.txt.

    @attention: .po file must have the same name as the Plugin's name property.

    @bug: I think there are packaging issues related with translation_path
    """

    __metaclass__ = classinit.ClassInitMeta
    __classinit__ = classinit.build_properties

    directory = ""

    translation_path = ""

    config = None

    components = {}
    name = ""
    plugin_dependencies = []
    external_dependencies = []
    version = ""
    description = ""

    def __init__(self):
        """
        Constructor
        """

        # configure the log category based on my name
        self.log_category = self.name

        log.Loggable.__init__(self)
        self.debug("Creating")

        # localization support
        self.install_localization()

    @classmethod
    def load_config(cls):
        components = {}

        if cls.config:
            plugin_dir = os.path.dirname(inspect.getsourcefile(cls))
            config_file = os.path.join(plugin_dir, cls.config)
            config = Config(config_file)
            general_section = config.get_section('general',{})

            name = general_section.get('name','')
            if name:
                cls.name = name

            version = general_section.get('version','')
            if version:
                # TODO: convert to a tuple?
                cls.version = version

            description = general_section.get('description','')
            if description:
                cls.description = description

            deps = general_section.get('plugin_dependencies',[])
            if deps:
                cls.plugin_dependencies = deps

            ext_deps = general_section.get('external_dependencies',[])
            cls.external_dependencies = ext_deps

            sections = config.as_dict()
            del sections['general']

            # scan components
            for component_path, section in sections.iteritems():
                if 'name' not in section:
                    path = component_path.split(':')
                    if len(path) < 2:
                        continue
                    path = path[1]
                    un_camel = misc.un_camelify(path)
                    section['name'] = un_camel
                component_name = section['name']
                section['path'] = component_path

                components.update({component_name:section})

        if components:
            cls.components = components

    @classmethod
    def initialize(cls):
        """
        Check Plugin's python dependencies and check each Component of
        the Plugin. Invalid components are removed from the Plugin.

        @raise InitializeFailure: if the Plugin is not well structured
        @raise UnMetDependency:   if the plugin has un-met dependencies
        @rtype:                   list of L{elisa.core.component.ComponentError}
        """
        log.debug(cls.name, "Initializing")
        errors = []
        invalid_components = []

        # check plugin's python dependencies
        component.check_python_dependencies(cls.name,
                                            cls.external_dependencies)

        if not isinstance(cls.components, dict):
            reason = "%s.components must be a dictionnary" % cls.__name__
            raise component.InitializeFailure(cls.name, reason)

        # check platform and python dependencies for each component
        for component_name, informations in cls.components.iteritems():
            platforms = informations.get('platforms',[])
            deps = informations.get('external_dependencies',[])

            try:
                component.check_platforms(component_name, platforms)
                component.check_python_dependencies(component_name, deps)
            except component.ComponentError, error:
                errors.append(error)
                invalid_components.append(component_name)

        # remove invalid components
        for component_name in invalid_components:
            del cls.components[component_name]

        return errors

##     def post_init(self):
##         # TODO: move that away
##         for table_name, sql in self.db_tables.iteritems():
##             query = "CREATE TABLE %s (%s);" % (table_name, sql)
##             self.sql_execute(query)


##     def sql_execute(self, query, log=False):
##         result = None
##         media_mgr = common.get_application().get_media_manager()
##         if media_mgr.enabled():
##             connection = media_mgr.get_db_connection()
##             result = connection.execute(query, log=log)
##         return result

    def install_localization(self):
        """
        Compile plugin's message translations using gettext.

        FIXME: that requires write permission access to the
        translation_path directory. Needs to be checked before any
        action
        """
        if self.translation_path:
            for path, names, filenames in os.walk(self.translation_path):
                for i, dirname in enumerate(names):
                    if dirname.startswith('.'):
                        del names[i]
                    for f in filenames:
                        if f.endswith('.po'):
                            src = os.path.join(path, f)
                            dest = src[:len(src) - 3] + '.mo'
                            # compile the .po if necessary
                            if not os.path.exists(dest):
                                msgfmt.make(src, dest)
                            else:
                                src_mtime = os.stat(src)[8]
                                dest_mtime = os.stat(dest)[8]
                                if src_mtime > dest_mtime:
                                    msgfmt.make(src, dest)



##     @classmethod
##     def enable(cls):
##         cls.logger.info('Enabling plugin %s' % cls.name)
##         app_config = common.get_application().get_config()
##         ep_name = cls.entry_point_name
##         option = app_config.get_option(ep_name)
##         if type(option) == list:
##             option.append(cls.name)
##         else:
##             option = cls.name
##         app_config.set_option(ep_name, option)

##     def disable(self):
##         self.logger.info('Disabling plugin %s' % self.name)
##         app_config = common.get_application().get_config()
##         ep_name = self.entry_point_name
##         option = app_config.get_option(ep_name)
##         if type(option) == list and self.name in option:
##             option.remove(self.name)
##         else:
##             option = ''
##         app_config.set_option(ep_name, option)
