# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

__maintainer__ = 'Lionel Martin <lionel@fluendo.com>'
__maintainer2__ = 'Florian Boucault <florian@fluendo.com>'

import threading

class ThreadsafeList(list):

    # FIXME: only some method add threadsafety

    # FIXME: if an exception is raised, the lock does not get released; in
    # order to do that, the release should be in a finally block; e.g.:
    # try:
    #     list.__delitem(self, rank)
    # finally:
    #     self._list_lock.release()

    def __init__(self, *args, **kw):
        self._list_lock = threading.Lock()
        list.__init__(self, *args, **kw)

    def __delitem__(self, rank):
        self._list_lock.acquire()
        list.__delitem__(self, rank)
        self._list_lock.release()

    def __eq__(self, other):
        self._list_lock.acquire()
        # FIXME: this changes the behaviour of python lists
        r = id(self) == id(other)
        self._list_lock.release()
        return r

    def __getitem__(self, rank):
        self._list_lock.acquire()
        r = list.__getitem__(self, rank)
        self._list_lock.release()
        return r

    def __delslice__(self, i, j):
        self._list_lock.acquire()
        list.__delslice__(self, i, j)
        self._list_lock.release()

    def __iadd__(self, elements):
        self._list_lock.acquire()
        ret = list.__iadd__(self, elements)
        self._list_lock.release()
        return ret

    def __imul__(self, mult):
        self._list_lock.acquire()
        ret = list.__imul__(self, mult)
        self._list_lock.release()
        return ret

    def __len__(self):
        self._list_lock.acquire()
        r = list.__len__(self)
        self._list_lock.release()
        return r

    def __setitem__(self, rank, value):
        self._list_lock.acquire()
        list.__setitem__(self, rank, value)
        self._list_lock.release()

    def __setslice__(self, i, j, sequence):
        self._list_lock.acquire()
        list.__setslice__(self, i, j, sequence)
        self._list_lock.release()

    def append(self, element):
        self._list_lock.acquire()
        list.append(self, element)
        self._list_lock.release()

    def count(self, element):
        self._list_lock.acquire()
        ret = list.count(self, element)
        self._list_lock.release()
        return ret

    def extend(self, elements):
        self._list_lock.acquire()
        list.extend(self, elements)
        self._list_lock.release()

    def index(self, element):
        self._list_lock.acquire()
        ret = list.index(self, element)
        self._list_lock.release()
        return ret

    def insert(self, position, element):
        self._list_lock.acquire()
        list.insert(self, position, element)
        self._list_lock.release()

    def pop(self, position=-1):
        self._list_lock.acquire()
        ret = list.pop(self, position)
        self._list_lock.release()
        return ret

    def remove(self, element):
        self._list_lock.acquire()
        list.remove(self, element)
        self._list_lock.release()

    def reverse(self):
        self._list_lock.acquire()
        list.reverse(self)
        self._list_lock.release()

    def sort(self, key=None):
        self._list_lock.acquire()
        list.sort(self,key=key)
        self._list_lock.release()
