# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


__maintainer__ = 'Philippe Normand <philippe@fluendo.com>'
__maintainer2__ = 'Florian Boucault <florian@fluendo.com>'

"""
Last reviewed: 07/06/12 by Benjamin
    - possible problem in FilterUriObserver._set_range
      which also could cause problems in FilterUriObserver.remove
      and FilterUriObserver.inserted
    - question in FilterUriObserver.process_element
    - cleans in MediaMenuActivity.set_directory_actions
    - problem in MediaMenuActivity.create_menu_for_uri
    - MediaMenuActivity.got_children asks for not existing variable
        or I don't know where it comes from..
    - MediaMenuActivity._handle_media_event couldn't be used as callback
      for media_manager.monitor_uri. Described there
"""

from menu_activity import MenuActivity
from elisa.core.observers.list import ListObserver, ListObservable
from elisa.core.observers.dict import DictObserver, DictObservable
from elisa.core.media_manager import MediaProviderNotFound
from elisa.core import common, log
from elisa.core import media_uri, player
from elisa.core.utils import misc
from elisa.core.bus import bus_message
from elisa.base_components.media_provider import NotifyEvent
from twisted.internet import defer

import os, re

class UriFilter(log.Loggable):
    log_category = 'uri_filter'
    
    def __init__(self, activity):
        log.Loggable.__init__(self)        
        self._activity = activity
        self.valid_media_types = ['directory',] + self._activity.media_types

    def _got_media_type(self, media_type, uri):
        ok =  media_type and media_type['file_type'] in self.valid_media_types

        if not ok:
            self._activity.debug("%s is an invalid media for this activity",
                                 uri)
            self._activity.debug("%r %r", media_type, self.valid_media_types)
        return ok
    
    def __call__(self, uri):
        try:
            dfr = common.application.media_manager.get_media_type(uri)
        except MediaProviderNotFound:
            self.warning("No MediaProvider found for %s" % uri)
            dfr = defer.Deferred()
            dfr.callback(False)
            return dfr
        dfr.addCallback(self._got_media_type, uri)
        return dfr

class MetadataObserver(DictObserver):

    def __init__(self, model, activity):
        DictObserver.__init__(self)
        self._activity = activity
        self._model = model

    def modified(self, key, value):
        self._activity.metadata_changed(self._model, key, value)

class FilteringUriObserver(ListObserver):

    def __init__(self, model, activity, start=0, end=None):
        ListObserver.__init__(self)
        self._model = model
        self._activity = activity
        self._set_range(start, end)
        self._uri_filter = UriFilter(self._activity)

    def _set_range(self, start, end):
        # Review: Shouldn't we check that start and end are at least
        # integers and of course that they are >= 0? or we may get
        # ugly things: self._model[-1] ...  it also courses to bad
        # stuff in inserted and in removed. In some cases it breaks!
        self._range = (start, end)

        self._model[0:start] = []
        if end != None:
            self._model[end-start:] = []

    def _filter_elements(self, elements):
        filtered_elts = []

        def filtered(r, elt):
            if r:
                filtered_elts.append(elt)

        dfrs = []

        for elt in elements:
            uri = elt[0]
            sdfr = self._uri_filter(uri)
            sdfr.addCallback(filtered, elt)
            dfrs.append(sdfr)

        dfr = defer.DeferredList(dfrs)
        dfr.addCallback(lambda r: filtered_elts)
        return dfr

    def refresh(self, elements, start, end):
        self._set_range(start, end)
        # FIXME: err this is ugly, dude

        dfr = self._filter_elements(elements)

        def final(filtered_elts):
            uris = [ m.uri for m in self._model if m.uri ]
            elements = filtered_elts
            if not len(elements):
                self._model.has_children = False
                self._model.loading = False

            for element in elements:
                if element[0] not in uris:
                    self._observable.append(element)
                uris = [ m.uri for m in self._model if m.uri ]
                
        dfr.addCallback(final)

    def inserted(self, elements, position):
        # Review: see _set_range
        r1, r2 = self._range

        dfr = self._filter_elements(elements)

        def filtered(filtered_elts):
            elements = filtered_elts
            if r2 != None:
                # cropping inserted elements depending on the range
                valid1 = r1 - position
                if valid1 < 0:
                    valid1 = 0
                valid2 = position + len(elements) - r2
                if valid2 <= 0:
                    elements = elements[valid1:]
                else:
                    elements = elements[valid1:-valid2]

                # cropping models to give place to the new elements
                max = r2 - r1
                nb_del = len(elements) + len(self._model) - max
                if nb_del > 0:
                    self._model[-nb_del:] = []
            else:
                # cropping inserted elements depending on the range
                valid1 = r1 - position
                if valid1 < 0:
                    valid1 = 0
                elements = elements[valid1:]

            models = []
            li = []
            for elt in elements:
                dfr2 = self.process_element(elt)
                dfr2.addCallback(lambda model: models.append(model))
                li.append(dfr2)

            dfr = defer.DeferredList(li)

            def final(r):
                for model in models:
                    if model is not None:
                        pos = self._activity.files_index_start
                        for node in self._model[self._activity.files_index_start:]:
                            if model < node:
                                break
                            pos +=1
                            
                        self._model.insert(pos, model)

            dfr.addCallback(final)
            return dfr

        dfr.addCallback(filtered)

    def removed(self, elements, position):
        # Review: see _set_range
        r1, r2 = self._range
        start = position - r1
        stop = start + len(elements)
        if start < 0:
            start = 0
        if stop < 0:
            stop = 0

        self._model[start:stop] = []

        if self._range[1]:
            # refill the model with elements from self._observable
            nb_selected = stop - start
            selected = self._observable[r2-nb_selected:r2]

            li = []
            for elt in selected:
                dfr2 = self.process_element(elt)
                dfr2.addCallback(lambda model: self._model.append(model))
                li.append(dfr2)

            dfr = defer.DeferredList(li)

            def final(r):
                pass

            dfr.addCallback(final)


    def process_element(self, element):
        uri = element[0]

        dfr = common.application.media_manager.get_media_type(uri)
        d = self._activity.create_menu_for_uri(element, dfr,
                                               parent=self._model,
                                               start_end=self._range)
        return d


class MediaMenuActivity(MenuActivity):
    """

    Notes:

     - MediaMenuActivities should declare in their default config a
       "locations" (type: list) option
     - All menu models created by this activity:

       * have their icon name prefixed with the media_type value and a dash

    Example: the "By folder" menu of the AudioActivity would have the
    "audio-by-folder" icon

    @note: MenuModels should be free-ed

    @cvar menu_name:      Media menu name, should be human readable
    @type menu_name:      string
    @cvar menu_icon_name: Menu icon name
    @type menu_icon_name: string
    @cvar media_types:    Media types handled by the activity
    @type media_type:     list
    @ivar root_menu:      MenuModel representing the menu tree of the activity
    @type root_menu:      L{elisa.plugins.base.models.menu_model.MenuModel}
    @ivar loadmore_callbacks: the loadmore callbacks for the menu_models handled by the activity
    @type loadmore_callbacks: dict mapping menu_models to callables
    """

    slideshow_model = None
    player_model = None

    menu_name = ""
    menu_icon_name = ""
    media_types = []
    xdg_var_name = ''

    def __init__(self):
        MenuActivity.__init__(self)
        self.files_index_start = 0
        self.root_menu = None
        self.loadmore_callbacks = {}
        self._media_locations = []
        self._hotplugged_devices = []
        self.model_cache = {}
        self.meta_observers = {}
        self.byfolders_cache = []

    def initialize(self):
        MenuActivity.initialize(self)
        bus = common.application.bus
        bus.register(self._got_bus_message, bus_message.MediaLocation)

        self.root_menu = MenuActivity._create_menu_model(self,
                                                         self.menu_name,
                                                         self.menu_icon_name)

        self._by_folder = self._create_menu_model("By Folder", "by_folder")
        self._by_location = self._create_menu_model("Network",
                                                    "browse_by_network")
        self.root_menu.append(self._by_folder)
        self.root_menu.append(self._by_location)
        self.root_menu.has_children = True


        self.loadmore_callbacks = {self._by_folder: self._loadmore_by_folder}

        self._load_media_locations()

    # Activity implementation

    def get_model(self):
        self.root_menu.activity = self
        return self.root_menu


    def set_directory_actions(self, menu_model, uri):
        """
        Add some menu_items at the beginning of the directory
        folder. Each of these menu_items can have customized activate
        actions.

        This method needs to update L{file_index_start} instance
        variable so that the activity knows when action items ends in
        the children list.

        @param menu_model: the directory model to add actions in
        @type menu_model: L{elisa.plugins.base.models.menu_node_model.MenuNodeModel}
        @param uri: directory URI
        @type uri: L{elisa.core.media_uri.MediaUri}
        """
        menu_model.uri = uri
        
        registry = common.application.plugin_registry
        action = registry.create_component('base:enqueue_action')
        action.uri = uri
        action.player_model = self.player_model
        
        action_model = self._create_menu_model('Play all', 'all')
        action_model.uri = uri
        action_model.activate_action = action
        menu_model.insert(0, action_model)

        self.files_index_start = 1
        
    def set_activate_action(self, menu_model, uri):
        """
        Add an activate action to menu items that are not directories.

        @param menu_model: the model to add an action in
        @type menu_model: L{elisa.plugins.base.models.menu_node_model.MenuNodeModel}
        @param uri: file URI
        @type uri: L{elisa.core.media_uri.MediaUri}
        """
        registry = common.application.plugin_registry
        action = registry.create_component('base:play_action')
        action.uri = uri
        action.player_model = self.player_model
        menu_model.activate_action = action

    # MenuActivity implementation

    def loadmore(self, model, start=0, end=None):
        dfr = None
        callback = self.loadmore_callbacks.get(model)
        if callback == None:
            callback = self._loadmore_for_model
        if callback != None:
            self.debug("Calling %r", callback)
            dfr = callback(model, start, end)

        return dfr

    def set_hover_action(self, menu_model, uri, file_type='directory'):
        if uri:
            registry = common.application.plugin_registry

            if file_type != 'directory':
                action_type = 'base:preview_play_action'

                action = registry.create_component(action_type)
                action.uri = uri
                action.player_model = self.player_model
                menu_model.hover_action = action

    def _create_menu_model(self, text, icon_name, thumbnail_source=None,
                           parent=None):
        if self.menu_icon_name:
            icon_name = "%s_%s" % (self.menu_icon_name, icon_name)
        return MenuActivity._create_menu_model(self, text, icon_name,
                                               thumbnail_source, parent)

    def create_menu_for_uri(self, uri_with_info, media_type, parent=None,
                            start_end=None, icon=None):
        """
        @param uri_with_info:   the URI and its optional metadata
        @type uri_with_info:    tuple (MediaUri, dict)
        """

        def create(media_type, uri_with_info,icon):
            uri = uri_with_info[0]
            metadata = uri_with_info[1]
            
            file_type = media_type['file_type']

            if not icon:
                if file_type == "directory":
                    icon = "directory"
                else:
                    icon = "file"

            if not metadata.has_key('default_image'):
                metadata['default_image'] = None

            if metadata['default_image'] is not None:
                thumbnail_source = metadata['default_image']
            elif file_type == "directory":
                thumbnail_source = None
            else:
                thumbnail_source = uri

            menu_model = self._create_menu_model(uri.label, icon,
                                                 thumbnail_source,
                                                 parent=parent)
            menu_model.uri = uri

            if file_type != 'directory':
                self.set_activate_action(menu_model, uri)

            self._observe_metadata_dict(menu_model, metadata)

            metadata_manager = common.application.metadata_manager
            metadata_manager.get_metadata(metadata)

            def update_model(r):
                menu_model.has_children = r
                return menu_model

            self.set_hover_action(menu_model, uri, file_type)
            supported_types = ['directory',] + self.media_types
            media_manager = common.application.media_manager

            dfr = media_manager.has_children_with_types(uri,
                                                        supported_types)
            dfr.addCallback(update_model)
            return dfr

        d = media_type
        d.addCallback(create, uri_with_info, icon)
        return d

    # Private methods

    def _got_bus_message(self, message, sender):

        def got_model(menu_model):
            menu_model.has_children = True

            self._hotplugged_devices.append(menu_model)
            if isinstance(message, bus_message.DeviceAction):
                self.root_menu.append(menu_model)
            else:
                self._by_location.append(menu_model)
                self._by_location.has_children = True

        if set(self.media_types).intersection(set(message.media_types)):
            uri = media_uri.MediaUri(message.mount_point)
            uri.label = message.name
            ActionType = bus_message.MediaLocation.ActionType
            if message.action == ActionType.LOCATION_ADDED:

                # no icons for DAAP and UPnP.. don't ask why
                if message.fstype in ('daap', 'upnp'):
                    icon = "computer"
                elif message.fstype == 'file':
                    # FIXME: hrmmmm
                    icon = "usb"
                else:
                    icon = message.fstype

                uri_with_info = (uri, {})
                try:
                    media_manager = common.application.media_manager
                    media_type = media_manager.get_media_type(uri)
                    dfr = self.create_menu_for_uri(uri_with_info,
                                                   media_type, icon=icon)
                except MediaProviderNotFound, error:
                    self.warning(error)
                else:
                    dfr.addCallback(got_model)
                    media_manager.add_source(uri, self.media_types)

            elif message.action == ActionType.LOCATION_REMOVED:
                if str(uri) in self.model_cache.keys():
                    del self.model_cache[str(uri)]
                menu_model = None
                for model in self._hotplugged_devices:
                    if model.uri == uri:
                        menu_model = model
                        self._hotplugged_devices.remove(menu_model)
                        for child in menu_model[:]:
                            menu_model.remove(child)
                        if isinstance(message, bus_message.DeviceAction):
                            self.root_menu.remove(menu_model)
                        else:
                            self._by_location.remove(menu_model)
                            if len(self._by_location) == 0:
                                self._by_location.has_children = False
                        break

    def _load_xdg_env(self):
        config_home = os.path.expanduser(os.getenv('XDG_CONFIG_HOME',
                                                   '~/.config'))
        user_dirs_file = os.path.join(config_home, 'user-dirs.dirs')
        if os.path.exists(user_dirs_file):
            xdg_enabled = len(filter(lambda i: i.startswith('XDG'),
                                     os.environ.keys())) > 0
            if not xdg_enabled:
                key_val_re = re.compile('(.*)="(.*)"')
                for line in open(user_dirs_file).readlines():
                    match = key_val_re.search(line)
                    if match and not match.groups()[0].startswith('#'):
                        var, value = match.groups()
                        value = misc.env_var_expand(value)
                        os.environ[var] = value
                        
    def _load_media_locations(self):
        self._load_xdg_env()
        
        # check XDG variables
        if self.xdg_var_name and self.xdg_var_name in os.environ:
            local_directory = os.environ[self.xdg_var_name]
            location_uri = media_uri.MediaUri('file://%s' % local_directory)
            self.add_location(location_uri)
        else:
            self.warning("XDG user-dir disabled")
            
        for location in self.config.get('locations', []):
            # TODO: handle /*

            location_options = self.config.get(location, {})

            if location.endswith('/'):
                location = location[:-1]
                
            location_uri = media_uri.MediaUri(location.decode('utf-8'))
            if 'label' in location_options:
                location_uri.label = location_options['label']
                
            self.add_location(location_uri)

        self._by_folder.has_children = len(self._media_locations) > 0

    def add_location(self, location_uri):
        media_manager = common.application.media_manager

        def got_result(is_directory):
            ok = True
            if not is_directory:
                handle = media_manager.blocking_open(location_uri)
                if handle is None:
                    ok = False
                else:
                    handle.blocking_close()
            if ok:
                media_manager.add_source(location_uri, self.media_types)
                self.info("Adding location: %r", location_uri)
                self._media_locations.append(location_uri)
                
        try:
            dfr = media_manager.is_directory(location_uri)
            dfr.addCallback(got_result)
        except MediaProviderNotFound, error:
            self.warning(error)
            
    def _start_monitoring(self, monitored_uri, children_with_info, model):
        model.has_children = len(children_with_info) > 0
        model.loading = False
        media_manager = common.application.media_manager
        if media_manager.uri_is_monitorable(monitored_uri):
            media_manager.monitor_uri(monitored_uri, self._handle_media_event,
                                      children_with_info)
        return children_with_info

    def _handle_media_event(self, uri, metadata, event, monitored_uri,
                            children):
        self.debug('Monitoring event received in %s : %s %s' % (monitored_uri,
                                                                uri, event))
        media_manager = common.application.media_manager

        # FIXME: hack 'cause "file:///uri/" != "file:///uri"
        parent_uri = media_uri.MediaUri(uri.parent[:-1])

        if parent_uri == monitored_uri:
            if event == NotifyEvent.ADDED:
                children.insert(0, (uri, metadata))
            elif event == NotifyEvent.REMOVED:
                found = None
                for infos in children:
                    if infos[0] == uri:
                        found = infos
                        break
                if found:
                    children.remove(found)
            if media_manager.enabled:
                media_manager.handle_notify_event(uri, event,
                                                  self.media_types)

    # loadmore specialized callbacks

    def _loadmore_search_by_folder(self, model, start=0, end=None):
        # TODO: implement
        pass

    def _loadmore_for_model(self, model, start, end):
        dfr = None
        uri = model.uri

        def got_children(children, uri, model, filter_uri):
            self._start_monitoring(uri, children, model)
            #filter_uri._set_range(self.files_index_start, end)
            #filter_uri.refresh(children, self.files_index_start, end)
            
        if uri:
            if str(uri) not in self.model_cache.keys():
                self.debug('Not in cache: %r', model)
                model.has_children = True
                children = ListObservable()
                filter_uri = FilteringUriObserver(model, self, 0, end)
                filter_uri.observe(children)

                self.model_cache[str(uri)] = (children, filter_uri)

                model.loading = True
                self.set_directory_actions(model, uri)
                media_manager = common.application.media_manager
                dfr = media_manager.get_direct_children(uri, children)
                if dfr != None:
                    dfr.addCallback(got_children, uri, model, filter_uri)
                else:
                    model.loading = False
                    model.has_children = False
            else:
                self.debug('In cache: %r' % model)
                dfr = defer.Deferred()
                children, filter_uri = self.model_cache[str(uri)]
                if not model.loading or len(children) > 0:
                    filter_uri.refresh(children, start, end)

                dfr.callback(children)

            if len(children) > 0:
                model.has_children = True
        else:
            self.debug("No URI for %r", model)
        return dfr

    def _loadmore_by_folder(self, model, start=0, end=None):
        children = ListObservable()
        filter_uri = FilteringUriObserver(model, self, start, end)
        filter_uri.observe(children)

        dfr = defer.Deferred()
        for location in self._media_locations:
            if location not in self.byfolders_cache:
                children.append((location, {}))
                self.byfolders_cache.append(location)

        dfr.callback(children)
        return dfr

    def metadata_changed(self, model, key, value):
        if key == 'default_image' and value is not None:
            model.thumbnail_source = value

    def _observe_metadata_dict(self, model, metadata):
        if isinstance(metadata, DictObservable):
            meta_observer = MetadataObserver(model, self)
            meta_observer.observe(metadata)
            self.meta_observers[model] = meta_observer
