# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

"""
This media provider provides access to the rss://-scheme and reads the
enclosures, if there are some

To use it do this:
get to your favorit podcast-website. copy the link of the rss, and paste it
into the locations of the audio (oder video for a video blog) activity. now
remove the http and replace it with rss:
http://www.rocketboom.com/vlog/quicktime_daily_enclosures.xml
 = rss://www.rocketboom.com/vlog/quicktime_daily_enclosures.xml

This URI is now accessable in your audio (video) activity inside browse by
folder.

enjoy it!


Troubleshooting:
You get no Children, but it is not loading anymore?
    1. Are you sure, that you entered it in the right activity? Podcasts are
    NOT shown in video activity and the other way round.
    2. Be sure that the rss you are using has really enclosures. For this, you
    can just read the file with your favorite editor and could search for
    'enclosure'. Only if they exist, we can access it...

If you are sure, that it should work, but it doesn't feel free to file a ticket
on the elisa trac. Don't forget to paste your link there!
"""


__maintainer__ = 'Benjamin Kampmann <benjamin@fluendo.com>'

from elisa.base_components.media_provider import MediaProvider
from elisa.core.media_uri import MediaUri
from elisa.core.media_file import MediaFile

from twisted.internet import defer, threads

import feedparser

class PodcatcherMedia(MediaProvider):
    """
    This reads the enclosures for an rss-scheme...
    """

    def __init__(self):
        MediaProvider.__init__(self)
        self._children = {}
        self._cached = []

    def scannable_uri_schemes__get(self):
        return {}

    def supported_uri_schemes__get(self):
        ### Feedparser also supports atom, so shouldn't we add it here?
        return { 'rss': 50 }

    def blocking_get_media_type(self, uri):
        media_type = {}
        if self.blocking_is_directory(uri):
            media_type = {'file_type': 'directory', 'mime_type': ''}
        else:
            media_type = self._children.get(uri, None)
        return media_type

    def blocking_is_directory(self, uri):
        return uri.host == 'CHILD'

    def blocking_has_children_with_types(self, uri, media_types):
        has_children = False
        if self._is_directory(uri):
            ## we need a kind of a MAYBE-Opertor. We don't know, if there
            ## are this children, before accessing...
            has_children = True
        return has_children
    
    def _retrieve_children(self, uri, list_of_children, add_info=False):
        """
        retrieve the children of uri and fills list

        @param uri:                     the URI to analyze
        @type uri:                      L{elisa.core.media_uri.MediaUri}
        @param list_of_children:        List where the children will be appended
        @type list_of_children:         list
        @param add_info:                add metadata information to the result
        @type add_info:                 bool
        """

        if self.blocking_is_directory(uri) and not uri in self._cached:
            real_http = "http%s" % str(uri)[3:]
            self.debug("Real URI is %s" % real_http)

            data = feedparser.parse(real_http)
            childs = []
            self._cached.append(uri)

            for entry in data['entries']:
                name = entry.title
                self.debug("Found entry %s" % name)
                mime = None
                try:
                    if entry.enclosures:
                        type = entry.enclosures[0]['type']
                        luri = MediaUri('rss://CHILD/%s' %
                                        entry.enclosures[0]['href'][7:])
                        luri.label = name
                        length = entry.enclosures[0]['length']
                        if type.find('audio') != -1:
                            mime = {'file_type' : 'audio', 
                                    'mime_type' : type}
                        elif type.find('video') != -1:
                            mime = {'file_type' : 'video', 
                                    'mime_type' : type}

                except AttributeError, e:
                    ### If there are no enclosures found...
                    pass

                ### We want only those, for which we know the mime!
                if mime: 
                    childs.append( luri )
                    self._children[luri] = mime

        for child in childs:
            list_of_children.append( (child, {}) )
        
        return list_of_children


    def get_direct_children(self, uri, l):
        d = threads.deferToThread(self._retrieve_children, uri,
                                  l, add_info=True)
        return d

    def get_real_uri(self, uri):
        return MediaUri("http:/%s" % uri.path)


if __name__ == "__main__":

    m = PodcatcherMedia()
    uri = MediaUri('rss://www.brennpunkt-ostkreuz.de/rssfeed/Brennpunktgeschichten.rss')
    m.has_children(uri)
    list =  []

    def printer (list):
        for i, meta in list:
            print "child", i, meta
            print "label", i.label
            print "real", m.get_real_uri(i)
        
    a = m.get_direct_children(uri, list)
    a.addCallback(printer)

    try:
        from twisted.internet import glib2reactor
        glib2reactor.install()
    except AssertionError:
        # already installed...
        pass

    from twisted.internet import reactor

    reactor.run()
