# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


__maintainer__ = 'Florian Boucault <florian@fluendo.com>'
__maintainer2__ = 'Philippe Normand <philippe@fluendo.com>'


from elisa.base_components.view import View
from elisa.core import common
from elisa.core import player
from twisted.internet import reactor
import gst

class PlayerView(View):
    """
    This class implements a view which can create a
    Player instance and manage it.
    This class can be inherited to provide the
    toolkit dependent display, as well as the
    GStreamer video and audio sinks
    """

    default_config = {'audio_sink_name': 'autoaudiosink',
                      'subtitle_font': 'Sans Bold 24'
                      }
    config_doc = {'audio_sink_name': 'Sink name the Player should use'}

    def __init__(self):
        View.__init__(self)
        self.player = common.application.player_controller.create_player()
        
    def initialize(self):
        View.initialize(self)
        sink = gst.element_factory_make(self.config['audio_sink_name'])
        self.player.audio_sink = sink
        self.player.subtitle_font = self.config['subtitle_font']
        self.player.end_of_stream.connect(self._end_of_stream)
        common.application.bus.register(self._is_playing,
                                        player.PlayingStarted)

    def _is_playing(self, message, sender):
        self.player.volume = self.controller.model.volume
        self.controller.model.file_type = self.player.file_type
        
    def _end_of_stream(self, uri):
        self.controller.model.state = player.STATES.STOPPED

    def attribute_set(self, key, value):
        self.debug("attribute %r set to %r", key, value)
        if key == 'state':
            if value == player.STATES.PLAYING:
                self.player.volume = self.controller.model.volume
                self.state_changed(value)
                self.player.play()
            elif value == player.STATES.PAUSED:
                self.state_changed(value)
                self.player.pause()
            elif value == player.STATES.STOPPED:
                self.state_changed(value)
                self.player.stop()
        elif key == 'uri':
            self.player.load_uri(value)
            self.uri_changed(value)
        elif key == 'seek_to':
            self.seek_to_changed(value)
        elif key == 'volume':
            self.volume_changed(value)
        elif key == 'muted':
            self.player.muted = value
        elif key == 'pause_requested':
            self.pause_requested_changed(value)
        elif key == 'focused':
            self.focus_changed(value)
            
    def state_changed(self, state):
        self.player.volume = self.controller.model.volume
        self.player.muted = self.controller.model.muted
        self.controller.model.duration = self.player.duration
        self._update_model_position()

    def _update_model_position(self):
        if self.controller.model.duration != self.player.duration:
            self.controller.model.duration = self.player.duration
        self.controller.model.position = self.player.position
        if self.controller.model.state == player.STATES.PLAYING:
            reactor.callLater(0.5, self._update_model_position )

    def uri_changed(self, uri):
        pass

    def pause_requested_changed(self, value):
        pass

    def focus_changed(self, value):
        pass

    def volume_changed(self, value):
        self.player.volume = value
        
    def seek_to_changed(self, position):
        self.player.seek_to_location(position)

    def controller_changed(self):
        if self.controller.uri != None:
            self.player.load_uri(self.controller.uri)

            state = self.controller.state
            if state == player.STATES.PLAYING:
                self.player.play()
            elif state == player.STATES.PAUSED:
                self.player.pause()
            elif state == player.STATES.STOPPED:
                self.player.stop()
