# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


__maintainer__ = 'Florian Boucault <florian@fluendo.com>'


from elisa.core import plugin_registry
from pypgmtools.graph.image import Image
from pypgmtools.graph.group import Group
from pypgmtools.widgets.player_osd import *
from pypgmtools.widgets.volume_osd import VolumeOsd

from twisted.internet import reactor
from elisa.core import player

import pygst
pygst.require('0.10')
import gst, pgm


PlayerViewClass = plugin_registry.get_component_class('base:player_view')

class PlayerView(PlayerViewClass):
    supported_controllers = ('base:player_controller')

    def __init__(self):
        PlayerViewClass.__init__(self)
        self.context_path = "pigment:pigment_context"
        self.background = Image()
        self.background.position = (0.0, 0.0, 0.0)
        self.background.size = (4.0, 3.0)
        self.background.bg_color = (0, 0, 0, 0)
        self.background.opacity = 255
        self.background.visible = True

        video_sink = gst.element_factory_make('pgmimagesink')
        video_sink.set_property('image', self.background)
        self.player.video_sink = video_sink
        self.player_dock = None
        self.group = None
        self.player.loading.connect(self.enable_loading_animation)
    
        
    def enable_loading_animation(self, enabled):
        if enabled and self.frontend:
            video_wait_icon = self.frontend.theme.get_media("video_wait_icon")
            self.background.set_from_fd(os.open(video_wait_icon, os.O_RDONLY))
        else:
            self.background.clear()
            


    def frontend_changed(self, previous_frontend, new_frontend):
        PlayerViewClass.frontend_changed(self, previous_frontend, new_frontend)

        if self.group == None:
            # create the group
            canvas = self.frontend.context.viewport_handle.get_canvas()
            self.group = Group(canvas, pgm.DRAWABLE_MIDDLE)
            self.context_handle = self.group

            dock_bg_path = self.frontend.theme.get_media("dock_background")
            dock_bar_bg_path = self.frontend.theme.get_media("dock_bar_bg")
            dock_bar_fg_path = self.frontend.theme.get_media("dock_bar_fg")

            px, py = 0.7, 0.8
            iw, ih = canvas.width, canvas.height

            self.player_dock = PlayerOsd(canvas,
                                         pgm.DRAWABLE_NEAR,
                                         10000,
                                         iw*px, ih*0.15,
                                         dock_bar_bg_path,
                                         dock_bar_fg_path,
                                         dock_bg_path)
            self.player_dock.position = ((1-px)*iw/2.0, ih*py, 0.0)
            self.player_dock.opacity = 0
            self.player_dock.visible = True
            
            vol_bg_path = self.frontend.theme.get_media("vol_dock_background")
            self.volume_dock = VolumeOsd(canvas,
                                         pgm.DRAWABLE_NEAR,
                                         "Volume: %s%%",
                                         iw*0.3, ih*0.1,
                                         vol_bg_path)
            self.volume_dock.position = (0.15, 0.10, 0)
            self.volume_dock.opacity = 0
            self.volume_dock.visible = True


            self.group.add(self.background)
            self.group.visible = True

    def uri_changed(self, uri):
        self.player_dock.title = uri.label

    def focus_changed(self, focus):
        if self.controller.state == player.STATES.PLAYING and not focus:
            self.player_dock.hide()
            self.volume_dock.hide()
                
    def state_changed(self, state):
        PlayerViewClass.state_changed(self, state)
        if self.controller.model.uri:
            if state == player.STATES.PAUSED:
                self.player_dock.show()
                self.group.visible = True
            elif state == player.STATES.STOPPED:
                self.group.visible = False
            else:
                self.group.visible = True
            # FIXME: this whole refresh_dock looks hackish
            reactor.callLater(0.05, self.refresh_dock)
        elif state == player.STATES.STOPPED:
            self.group.visible = False
            
    def seek_to_changed(self, position):
        PlayerViewClass.seek_to_changed(self, position)
        self.player_dock.time = self.player.position / 1000000.00
        self.player_dock.show()

    def volume_changed(self, value):
        PlayerViewClass.volume_changed(self, value)
        self.volume_dock.volume = value * 10.
        if self.controller.state == player.STATES.PLAYING and \
               self.controller.focused:
            self.volume_dock.show()
            
    def pause_requested_changed(self, value):
        # show the dock if the user press OK one time
        if self.controller.model.state == player.STATES.PLAYING and value:
            if not self.player_dock.is_visible():
                self.player_dock.show()
                self.volume_dock.show()
            else:
                self.controller.model.state = player.STATES.PAUSED
                
    def refresh_dock(self):
        # FIXME: a division in a code called every 50 ms is not good
        duration = int(self.player.duration / 1000000)
        self.player_dock.playing_length = duration

        self.player_dock.time = self.player.position / 1000000.0

        if self.controller.state == player.STATES.PLAYING:
            reactor.callLater(0.05, self.refresh_dock)
