# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


__maintainer__ = 'Benjamin Kampmann <benjamin@fluendo.com>'

from elisa.base_components.media_provider import MediaProvider
from elisa.core.media_uri import MediaUri
from elisa.core import common, component
from elisa.core.bus import bus_message

from twisted.internet import defer, threads
from flickrest import Flickr, FlickrError
import os
from twill import commands as twill
#import thread

class FlickrMedia(MediaProvider):
    """
    DOCME
    """
    secret = '14693e87a1349c20'
    key = '4bb9b930c2f01fab6b3fd20e02d165fb'

    config_doc = {'per_page' :  'show so many entries should be' \
                                'shown per page?',
                  'login': 'Your Flickr username',
                  'password': 'Your Flickr password'
                  }


    default_config = {'per_page': '30',
                      'login': '',
                      'password': ''
                      }


    def initialize(self):
        MediaProvider.initialize(self)

        self._login = self.config.get('login','')
        self._password = self.config.get('password','')
        
        if not self._login or not self._password:
            msg = "Please configure the Flickr plugin"
            raise component.InitializeFailure(self.name, msg)

 
        self._per_page = int(self.config.get('per_page', '30'))
        self._show_pages = False # ## Hardcoded, but I'll be a config-option
                                 # ## when it is usable
        self._connected = False

        try:
            self._flickr = Flickr(self.key, self.secret, "write")
            threads.deferToThread(self._connect_flickr)
        except FlickrError, e:
            raise component.InitializeFailure(self.name, e.get('msg'))


    def _connected_cb(self, auth):
        self.debug("Connected, %r", auth)
        self._connected = True

        uri = "flickr:///"
        action_type = bus_message.MediaLocation.ActionType.LOCATION_ADDED
        msg = bus_message.MediaLocation(action_type, "Flickr", 'flickr', uri,
                                        media_types=['image',])
        common.application.bus.send_message(msg)

    def _error_cb(self, failure):
        self.warning('Failed to connect to flickr. Maybe the API-Key or the'
                     ' Secret is wrong. Error message :%s' % 
                                                    failure.getErrorMessage())

    def _connect_flickr(self):

        def auth_open_url(state):
            if state is None:
                self._connected_cb(state)
            else:
                twill.go(state['url'])
                twill.formvalue(1, 'login', self._login)
                twill.formvalue(1, 'passwd', self._password)
                twill.submit()
                twill.config('readonly_controls_writeable', 'True')
                try:
                    twill.formvalue(3, 'done_auth', '1')
                    twill.submit()
                except:
                    # already granted
                    pass
                dfr = self._flickr.authenticate_2(state)
                dfr.addCallbacks(self._connected_cb, self._error_cb)
                return dfr

        self.debug("Authenticating")
        self._flickr.authenticate_1().addCallbacks(auth_open_url,
                                                   self._error_cb)

    def scannable_uri_schemes__get(self):
        return []

    def supported_uri_schemes__get(self):
        return {'flickr' : 0}

    def get_real_uri(self, uri):
        if self.blocking_is_directory(uri):
            return uri

        server = uri.get_param('server')
        secret = uri.get_param('secret')
        id = uri.path
        # build the real URL for the Image
        return MediaUri(self._build_uri(id, server, secret))

    def _build_uri(self, id, server, secret):
        # TODO: add support for different sizes...
        return "http://static.flickr.com/%s/%s_%s_m.jpg" % (server, id, secret)

    def get_media_type(self, uri):
        df = threads.deferToThread(self.blocking_get_media_type, uri)
        return df

    def blocking_get_media_type(self, uri):
        if self.blocking_is_directory(uri):
            return {'file_type' : 'directory', 'mime_type' : ''}
        else:
            return {'file_type' : 'image', 'mime_type' : 'image/jpeg'}


    def is_directory(self, uri):
        df = threads.deferToThread(self.blocking_is_directory, uri)
        return df

    def blocking_is_directory(self, uri):
        return uri.host != "photo"

    def has_children_with_types(self, uri, media_types):
        df = threads.deferToThread(self.blocking_has_children_with_types, uri,
                                   media_types)
        return df

    def blocking_has_children_with_types(self, uri, media_types):
        has_children = False
        media_type_ok = len(set(['picture',
                                 'directory']).intersection(media_types)) > 0
        if uri.host != 'photo' and media_type_ok:
            has_children = True
        return has_children


    def get_direct_children(self, uri, children_with_info):
        if not self._connected:
            dfr = defer.Deferred()
            dfr.callback(children_with_info)

        elif uri.host == '':
            dfr = threads.deferToThread(self._build_main_menu, children_with_info)
        else:
            dfr = self._dispatch(uri, children_with_info)

        return dfr

    def _dispatch(self, uri, children_with_info):
        page = int(uri.get_param('page', '0'))

        if uri.host == 'mine':
            if uri.path[1:] == 'latest':
                dfr = self._flickr.photos_search(user_id='me', page=page,
                                                 per_page=self._per_page)
                dfr.addCallback(self._got_images, children_with_info, str(uri))

            elif uri.path[1:] == 'photosets':
                dfr = self._flickr.photosets_getList(per_page=self._per_page)
                dfr.addCallback(self._got_photosets, children_with_info)

            elif uri.path[1:] == "tags":
                dfr = self._flickr.tags_getListUser(per_page=self._per_page)
                dfr.addCallback(self._got_whotags, children_with_info)

            elif uri.path[1:] == 'favorites':
                dfr = self._flickr.favorites_getList(page=page)
                dfr.addCallback(self._got_images, children_with_info,str(uri))

            else:
                dfr = threads.deferToThread(self._build_mine_menu,children_with_info)

        elif uri.host == 'photo':
            dfr = defer.Deferred()
            dfr.callback(children_with_info)

        elif uri.host == 'photoset':
            dfr = self._flickr.photosets_getPhotos(photoset_id=uri.path[1:],
                                                   page=page,
                                                   per_page=self._per_page)
            dfr.addCallback(self._got_images_in_photoset, children_with_info,
                            str(uri))

        elif uri.host == 'interesting':
            dfr = self._flickr.interestingness_getList(page=page,
                                                       per_page=self._per_page)
            dfr.addCallback(self._got_images, children_with_info, str(uri))

        elif uri.host == "hottags":
            dfr = self._flickr.tags_getHotList()
            dfr.addCallback(self._got_hottags, children_with_info)

        elif uri.host == "tag":
            who = uri.get_param('who')
            dfr = self._flickr.photos_search(tags=uri.path[1:],
                                             user_id=who,
                                             page=page,
                                             per_page=self._per_page)
            dfr.addCallback(self._got_images, children_with_info, str(uri))

        return dfr

    def _build_mine_menu(self, list_of_children):

        # TODO: i18n me
        mine_menu  = {"flickr://mine/favorites" : "my Favorites",
                      "flickr://mine/latest" : "latest",
                      "flickr://mine/photosets" : "in photosets",
                      "flickr://mine/tags" : "by tag"}

        return self._from_list(list_of_children, mine_menu)


    def _build_main_menu(self, list_of_children):


        # TODO: i18n me
        main_menu = {"flickr://interesting/": "Interesting",
                     "flickr://hottags/": "Hot Tags",
#                     "flickr://mine/my Stuff"]
                     "flickr://mine/latest": "My Photos"}
        return self._from_list(list_of_children, main_menu)

    def _from_list(self, list_of_children, menu):

        for entry, label in menu.items():
            list_of_children.append((self._make_uri(entry, label), {}))


        return list_of_children

    def _make_uri(self, uri, label=None):
        m = MediaUri(uri)
        if label != None:
            m.label = label
        return m

    def _got_images(self, ret, list_of_children, url=""):
        photos = ret.find('photos')
        pages = int(photos.get('pages'))
        page = int(photos.get('page'))

        for nu in ret.findall("photos/photo"):
            self._add_photo(nu, list_of_children)

        # We don't support more pages yet
        if (pages > 1 and page == 1 and self._show_pages):
            i = 1
            while (i < pages):
                i = i + 1
                m = MediaUri(url)
                m.set_param('page', i)
                m.label = 'page %s' % i
                list_of_children.append((m,{}))
        return list_of_children

    def _got_images_in_photoset(self, ret, list_of_children, url=''):
        photos = ret.find('photoset')
        pages = int(photos.get('pages'))
        page = int(photos.get('page'))

        for nu in ret.findall("photoset/photo"):
            self._add_photo(nu, list_of_children)

        #  We don't support more pages before the GUADEC 07
        if (pages > 1 and page == 1 and self._show_pages):
            i = 1
            while (i < pages):
                i = i + 1
                m = MediaUri(url)
                m.set_param('page', i)
                m.label = 'page %s' % i
                list_of_children.append((m,{}))
        return list_of_children

    def _add_photo(self, nu, list_of_children):
        id = nu.get('id')
        title = nu.get('title')
        server = nu.get('server')
        secret = nu.get('secret')

        uri = self._make_uri("flickr://photo/%s?server=%s&secret=%s" %
                                    (id, server, secret), title)
        list_of_children.append((uri, {
                'title' : title,
                'server' : server,
                'secret' : secret,
                'owner' : nu.get('owner'),
                'ispublic' : nu.get('ispublic'),
                'isfriend' : nu.get('isfriend'),
                'isfamily' : nu.get('isfamily'),
                'default_image': MediaUri(self._build_uri(id,server, secret))
                                        }
                                    )
                                )
        return list_of_children

    def _got_hottags(self, ret, list_of_children):
        for nu in ret.findall("hottags/tag"):
            name = nu.text
            score = nu.get('score')
            uri = MediaUri("flickr://tag/%s" % name)
            list_of_children.append(( uri, {"name" : name, score: "score"}))
        return list_of_children

    def _got_whotags(self, ret, list_of_children):
        who = ret.find("who").get('id')
        for nu in ret.findall("who/tags/tag"):
            name = nu.text
            score = nu.get('score')
            uri = MediaUri("flickr://tag/%s?who=%s" % (name, who))
            list_of_children.append(( uri, {"name" : name, score: "score"}))
        return list_of_children

    def _got_photosets(self, ret, list_of_children):
        for nu in ret.findall("photosets/photoset"):
            id = nu.get('id')
            name = nu.find('title').text
            server = nu.get('server')
            photos = nu.get('photos')
            secret = nu.get('secret')
            primary = nu.get('primary')
            uri = self._make_uri("flickr://photoset/%s" % id, name)
            list_of_children.append((uri, { 'name' : name,
                                            'id' : id,
                                            'server' : server,
                                            'photos' : photos,
                                            'secret' : secret,
                                            'primary' : primary}
                                            )
                                    )
        return list_of_children
