# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


__maintainer__ = 'Lionel Martin <lionel@fluendo.com>'
__maintainer2__ = 'Florian Boucault <florian@fluendo.com>'


from elisa.base_components.context import Context
from elisa.core.utils import signal

import pgm

from twisted.internet import reactor

class PigmentContext(Context):

    default_config = {'screen_ratio' : 'auto', 'window_width': '0'}
    config_doc = {'screen_ratio' : 'The ratio of the screen. This is a string'
                                    ' containing a relation value, seperated'
                                    ' by a colon (:). Common values are 16:9,'
                                    ' 4:3, 16:10, but it could be any other,'
                                    ' too. You can also enter the special'
                                    ' value mm, which uses the milimeters for'
                                    ' autodetection. Any other values without'
                                    ' a colon lead to the resolution base'
                                    ' autodetection!',
                  'window_width' : 'Here you can set the width in pixels the'
                                    ' window should have at startup. The heigth'
                                    ' is set relativ using screen_ratio.'
                                    ' If this value is 0,  we decide automatically.'
    }


    def initialize(self):
        Context.initialize(self)
        ratio = self.config.get('screen_ratio', 'auto')
        config_width = self.config.get('window_width', '0')
            
        # OpenGL viewport creation
        factory = pgm.ViewportFactory('opengl')
        viewport = factory.create()
        self.viewport_handle = viewport
        self._original_screen_size_mm = viewport.screen_size_mm
        viewport.title = 'Elisa Media Center'

        try:
            width = int(config_width)
            if width != 0:
               viewport.width = width
        except Exception, e:
            self.warning("The configuration value of window_width seems to be"
                         " no number value in pixels: %s" % e)

        # Canvas creation
        self.canvas = pgm.Canvas()
        # the input events are supplied by the viewport
        self.input_events_handle = viewport
        
        self.aspect_ratio = ratio
        
        # set the canvas aspect ratio to the screen aspect ratio
        w,h = self._get_screen_size_from_ratio()
        screen_ratio = w / h
        self.canvas.height = self.canvas.width/screen_ratio
        # Bind the canvas to the OpenGL viewport
        viewport.set_canvas(self.canvas)

        # show the window
        viewport.show()

        # delay the resize of the canvas to avoid crippling the CPU
        self._resize_delay = 0.200
        self._resize_delayed = None
        # signal triggered when the canvas gets resized
        self.canvas_resized = signal.Signal('canvas_resized', tuple)
        viewport.connect('configure-event', self._configure_callback)

        # delay the hiding of the mouse cursor
        self._cursor_delay = 1.000
        self._cursor_delayed = None
        self.viewport_handle.cursor = pgm.VIEWPORT_NONE
        # signal triggered when the canvas gets resized
        viewport.connect('motion-notify-event', self._motion_notify_callback)

    
    def _get_screen_size_from_ratio(self):
        # retrieve the screen aspect ratio
        if self.aspect_ratio == 'mm':
            self.debug("Using the milimeter based autodetection")
            w, h = self._original_screen_size_mm 
        elif self.aspect_ratio.find(':') != -1:
            w, h = self.aspect_ratio.split(':')
        else:
            w, h = self.viewport_handle.screen_resolution
            
        return float(w), float(h)
        

    def aspect_ratio_changed(self, aspect_ratio):
        w,h = self._get_screen_size_from_ratio()
        screen_ratio = w / float(h)
                
        self.debug("Screen ratio is %d" % screen_ratio)
        # set the window aspect ratio to the screen aspect ratio
        self.viewport_handle.height = self.viewport_handle.width/screen_ratio
        w2,l2 = self.viewport_handle.size
        self.viewport_handle.screen_size_mm = (int(w), int(h) )
        self.viewport_size = self.viewport_handle.size
        
        self.config['screen_ratio'] = aspect_ratio
        
    def _configure_callback(self, viewport, event):
        # (re)schedule the resize of the canvas
        if self._resize_delayed != None and self._resize_delayed.active():
            self._resize_delayed.reset(self._resize_delay)
        else:
            self._resize_delayed = reactor.callLater(self._resize_delay, self._resize_canvas)

    def _resize_canvas(self):
        # only reacts if the size of the viewport changed
        if self.viewport_size != self.viewport_handle.size:
            self.debug("Regeneration of the drawables")
            self.viewport_size = self.viewport_handle.size
            self.canvas_resized.emit(self.canvas.size)

        """
        # resize the canvas to follow the aspect ratio of the viewport
        viewport_ratio = float(self.viewport_handle.width)/float(self.viewport_handle.height)
        canvas_ratio = float(self.canvas.width)/float(self.canvas.height)

        # only resize the canvas if the aspect ratio of the viewport has changed
        if viewport_ratio != canvas_ratio:
            # FIXME: only one of width/height needs to change (multiplied by
            # viewport_ratio)
            self.canvas.size = (self.viewport_handle.width/100.0,
                                self.viewport_handle.height/100.0)
            self.debug("The canvas is getting resized. New size = (%s, %s)" % \
                        (self.canvas.size[0], self.canvas.size[1]))
        """

    def _motion_notify_callback(self, viewport, event):
        # show the cursor
        viewport.cursor = pgm.VIEWPORT_INHERIT

        # (re)schedule the hiding of the cursor
        if self._cursor_delayed != None and self._cursor_delayed.active():
            self._cursor_delayed.reset(self._cursor_delay)
        else:
            self._cursor_delayed = reactor.callLater(self._cursor_delay, self._hide_cursor)

    def _hide_cursor(self):
        self.viewport_handle.cursor = pgm.VIEWPORT_NONE

    def update(self):
        self.viewport_handle.update()

    def context_handle__set(self, context_handle):
        self._context_handle = context_handle
        if isinstance(context_handle, pgm.Drawable):
            self.canvas.add(pgm.DRAWABLE_MIDDLE, context_handle)

    def context_handle__get(self):
        return self._context_handle

    def clean(self):
        # FIXME: should unref self.viewport_handle
        pass
#        self.viewport_handle.setcanvas(None)
#        print len(gc.get_referrers(self.viewport_handle))
#        del self.viewport_handle
#        gc.collect()
