# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.


__maintainer__ = 'Florian Boucault <florian@fluendo.com>'


from elisa.core import plugin_registry
TreeViewClass = plugin_registry.get_component_class('base:tree_view')

import pgm
from pypgmtools.widgets.top_level_menu import *
from pypgmtools.timing import implicit
from pypgmtools.graph.group import Group
from pypgmtools.graph.text import Text

from twisted.internet import reactor


class TreeView(TreeViewClass):

    supported_controllers = ('poblenou:tree_controller',)
    default_associations = (
    ('poblenou:node_controller', 'poblenou:node_view'),
    )

    def __init__(self):
        TreeViewClass.__init__(self)
        self.context_path = 'pigment:pigment_context'

        # description widget
        self._description = None
        self._description_string = ""
        # the descriptive string of the current selected item is displayed
        # after a given time
        self._description_delay = 0.200
        self._description_delayed = None

        # menu widget
        self._menu = None

        # arrow widget that indicates that an item contains more items
        self._arrow = None
        self._arrow_delay = 0.200
        self._arrow_delayed = None


    def frontend_changed(self, previous_frontend, new_frontend):
        # FIXME: re-bind the widgets to the new canvas

        # connects to the canvas_resized signal of the new context
        if new_frontend != previous_frontend:
            new_frontend.context.canvas_resized.connect(self.canvas_resized)
            new_frontend.theme_changed.connect(self.theme_changed)

    def canvas_resized(self, size):
        canvas = self.frontend.context.canvas

        if self._description != None:
            self._scale_description()
        if self._arrow != None:
            self._scale_arrow()
        if self._menu != None:
            self._scale_menu()
            
    def theme_changed(self, new_theme):
        index = 0
        for child_controller in self.controller:
            icon, blurred, reflected = self._get_child_icons(child_controller)
            self._menu.update_images(index, icon, blurred, reflected)
            index += 1
        
    def controller_changed(self):
        canvas = self.frontend.context.canvas
        self.root_group = Group(canvas, pgm.DRAWABLE_MIDDLE)
        self.context_handle = self.root_group

        self._create_menu()
        self._create_description()
        self._create_arrow()

        TreeViewClass.controller_changed(self)

    def _create_menu(self):
        canvas = self.frontend.context.canvas
        # FIXME: move these 2 to canvas_resized
        menu_selected_position = (canvas.width*0.475, canvas.height*(-0.13), -10.0)
        menu_selected_size = (canvas.width, canvas.width)

        self._menu = TopLevelMenu(canvas, pgm.DRAWABLE_MIDDLE,
                                  width                 = canvas.width,
                                  height                = canvas.height,
                                  font_family           = "MgOpen Cosmetica",
                                  selected_position     = menu_selected_position,
                                  selected_size         = menu_selected_size,
                                  selected_opacity      = 32,
                                  mode                  = CAROUSEL,
                                  path_steps            = 1,
                                  duration              = 500,
                                  transformation        = implicit.DECELERATE)

        self._menu.bg_color = (0, 0, 0, 0)
        self._menu.opacity = 255
        self.root_group.add(self._menu)

        self._animated_menu = implicit.AnimatedObject(self._menu)
        self._animated_menu.mode = implicit.REPLACE
        self._animated_menu.setup_next_animations(duration = 550,
                                                  transformation = implicit.DECELERATE)
        self._scale_menu()
        self._menu.visible = True

    def _scale_menu(self):
        # set canvas size dependent properties of menu (size, position)
        canvas = self.frontend.context.canvas
        self._menu.width = canvas.width
        self._menu.height = canvas.height
 
        self._unselected_menu_position = ((canvas.width-self._menu.width)/2.0,
                                          (canvas.height-self._menu.height)/2.0,
                                          -500.0)
        self._selected_menu_position = ((canvas.width-self._menu.width)/2.0,
                                        (canvas.height-self._menu.height)/2.0,
                                         0.0)
        
        self._animated_menu.stop_animations()
        if self.controller.selected:
            self._menu.position = self._selected_menu_position
        else:
            self._menu.position = self._unselected_menu_position

        self._menu.layout()
        self._menu.update()
      
    def _create_description(self):
        # description text
        self._description = Text()
        self._description.label = ""
        self._description.font_family = "MgOpen Cosmetica"
        self._description.bg_color = (0, 0, 0, 0)
        self._description.alignment = pgm.TEXT_ALIGN_CENTER
        self._description.ellipsize = pgm.TEXT_ELLIPSIZE_END
        self._description.opacity = 255

        self._animated_description = implicit.AnimatedObject(self._description)
        self._animated_description.mode = implicit.REPLACE
        self._animated_description.setup_next_animations(duration = 400)

        self._scale_description()
        self._description.visible = True

        self.root_group.add(self._description)

    def _scale_description(self):
        # set canvas size dependent properties of description (size, position,
        # font_height)
        canvas = self.frontend.context.canvas
        self._description.font_height = canvas.height*0.073
        self._description.size = (canvas.width/2.0, canvas.height*0.095)
        self._description.x = (canvas.width - self._description.width)/2.0
        self._description.y = canvas.height * (1.0 - 0.023) - self._description.height
        self._description.z = 0.0

    def display_description(self, description):
        # hide the currently displayed description
        if self._description.opacity != 0:
            self._animated_description.stop_animations()
            self._description.opacity = 0

        self._description_string = description

        # (re)schedule the display of description
        if self._description_delayed != None and self._description_delayed.active():
            self._description_delayed.reset(self._description_delay)
        else:
            self._description_delayed = reactor.callLater(self._description_delay, self._show_description)

    def _show_description(self):
        # show the loaded description
        self._description_delayed = None
        self._description.label = self._description_string
        self._animated_description.opacity = 255

    def _create_arrow(self):
        # arrow indicating that more items are in the current selected item
        self._arrow = Image()
        self._arrow.bg_color = (0, 0, 0, 0)
        self._arrow.layout = pgm.IMAGE_SCALED
        self._arrow.alignment = pgm.IMAGE_CENTER
        self._arrow.opacity = 0

        arrow_path = self.frontend.theme.get_media("down_arrow_icon")
        self._arrow.set_from_fd(os.open(arrow_path, os.O_RDONLY))

        self._animated_arrow = implicit.AnimatedObject(self._arrow)
        self._animated_arrow.setup_next_animations(duration = 350)
        self._animated_arrow.setup_next_animations(attribute = "y",
                                                     duration = 350,
                                                     repeat_count = implicit.INFINITE,
                                                     repeat_behavior = implicit.REVERSE,
                                                     transformation = implicit.SMOOTH)
        self._scale_arrow()
        self._arrow.visible = True
        self.root_group.add(self._arrow)

    def _scale_arrow(self):
        canvas = self.frontend.context.canvas
        self._arrow.font_height = canvas.height*0.073
        # FIXME: height and y
        self._arrow.size = (canvas.width, canvas.height*0.03)
        self._arrow.x = (canvas.width - self._arrow.width)/2.0
        self._arrow.y = canvas.height * (1.0 - 0.01) - self._arrow.height
        self._arrow.z = 0.0
       
    def display_arrow(self, visible):
        # hide the arrow
        if self._arrow.opacity != 0:
            self._animated_arrow.stop_animations()
            self._arrow.opacity = 0
            self._scale_arrow()

        # (re)schedule the display of the arrow
        if self._arrow_delayed != None and self._arrow_delayed.active():
            if visible:
                self._arrow_delayed.reset(self._arrow_delay)
            else:
                self._arrow_delayed.cancel()
        elif visible:
                self._arrow_delayed = reactor.callLater(self._arrow_delay, self._show_arrow)

    def _show_arrow(self):
        self._animated_arrow.opacity = 255
        self._animated_arrow.y += 0.01

    def attribute_set(self, key, value):
        if key == 'current_index':
            self.debug("new index for %s: %s" % (id(self), value))
            self._menu.focused_item = value

        elif key == 'selected':
            self.debug("%s (un)selected: %s" % (id(self), value))
            if value:
                # display the menu
                if self._menu.is_selected():
                    self._menu.select()
                self._animated_menu.position = self._selected_menu_position

                # empty the description
                self.display_description("")
                self.display_arrow(False)

                # selected child
                child_index = self.controller.current_index
                if child_index >= 0 and child_index < len(self):
                    self[child_index].hide()

            else:
                # hide the menu
                if not self._menu.is_selected():
                    self._menu.select()
                self._animated_menu.position = self._unselected_menu_position
        
    def _get_child_icons(self, child_controller):
        # retrieve the paths to the icon, its blurred version and its reflected
        # version
        icon = child_controller.model.theme_icon
        icon_path = self.frontend.theme.get_media(icon)

        blurred_icon = icon[0:icon.find('_')] + "_blur_icon"
        blurred_icon_path = self.frontend.theme.get_media(blurred_icon)

        reflected_icon = icon[0:icon.find('_')] + "_reflected_icon"
        reflected_icon_path = self.frontend.theme.get_media(reflected_icon)

        return (icon_path, blurred_icon_path, reflected_icon_path)

    def child_view_creating(self, view, controller, position):
        TreeViewClass.child_view_creating(self, view, controller, position)

        # add a child view representation to the menu
        icon, blurred, reflected = self._get_child_icons(controller)
        self._menu.insert(position, icon, blurred, reflected, controller.model.text)
        self._menu.layout()
        self._menu.update()
